/*
 * Copyright (C) 2021 European Spallation Source ERIC.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package org.openepics.names.repository;

import java.util.ArrayList;
import java.util.List;

import javax.persistence.EntityManager;
import javax.persistence.PersistenceContext;
import javax.persistence.TypedQuery;
import javax.persistence.criteria.CriteriaBuilder;
import javax.persistence.criteria.CriteriaQuery;
import javax.persistence.criteria.Predicate;
import javax.persistence.criteria.Root;

import org.apache.commons.lang3.BooleanUtils;
import org.apache.commons.lang3.StringUtils;
import org.openepics.names.repository.model.Name;
import org.openepics.names.repository.model.NameStructure;
import org.openepics.names.rest.beans.FieldName;
import org.openepics.names.util.RepositoryUtil;
import org.springframework.stereotype.Repository;

/**
 * Handle name information in JPA.
 *
 * @author Lars Johansson
 */
@Repository
public class NameRepository {

    @PersistenceContext
    private EntityManager em;

    /**
     * Count names.
     *
     * @param deleted deleted
     * @param uuid uuid
     * @param name name
     * @param nameEquivalence name equivalence
     * @param systemStructure system structure mnemonic
     * @param deviceStructure device structure mnemonic
     * @param index index
     * @param description description
     * @return count of names
     */
    public Long countNames(Boolean deleted,
            String uuid, String name, String nameEquivalence, String systemStructure, String deviceStructure, String index, String description) {
        return countNames(deleted,
                uuid, name, nameEquivalence, systemStructure, deviceStructure, index, description,
                Boolean.FALSE);
    }

    /**
     * Count names.
     *
     * @param deleted deleted
     * @param uuid uuid
     * @param name name
     * @param nameEquivalence name equivalence
     * @param systemStructure system structure mnemonic
     * @param deviceStructure device structure mnemonic
     * @param index index
     * @param description description
     * @param includeHistory include history
     * @return count of names
     */
    public Long countNames(Boolean deleted,
            String uuid, String name, String nameEquivalence, String systemStructure, String deviceStructure, String index, String description,
            Boolean includeHistory) {

        // note
        //     use of function for mnemonic path
        // where
        //     latest, deleted
        //     queryFields, queryValues

        CriteriaBuilder cb = em.getCriteriaBuilder();
        CriteriaQuery<Long> cq = cb.createQuery(Long.class);
        Root<Name> from = cq.from(Name.class);

        cq.where(cb.and(preparePredicatesNames(cb, from, deleted,
                uuid, name, nameEquivalence, systemStructure, deviceStructure, index, description,
                includeHistory).toArray(new Predicate[0])));
        cq.select(cb.count(from));

        return em.createQuery(cq).getSingleResult();
    }

    /**
     * Find names.
     *
     * @param deleted deleted
     * @param uuid uuid
     * @param name name
     * @param nameEquivalence name equivalence
     * @param systemStructure system structure mnemonic
     * @param deviceStructure device structure mnemonic
     * @param index index
     * @param description description
     * @return list of names
     */
    public List<Name> readNames(Boolean deleted,
            String uuid, String name, String nameEquivalence, String systemStructure, String deviceStructure, String index, String description) {

        return readNames(deleted,
                uuid, name, nameEquivalence, systemStructure, deviceStructure, index, description,
                Boolean.FALSE, null, null, null, null);
    }

    /**
     * Find names.
     *
     * @param deleted deleted
     * @param uuid uuid
     * @param name name
     * @param nameEquivalence name equivalence
     * @param systemStructure system structure mnemonic
     * @param deviceStructure device structure mnemonic
     * @param index index
     * @param description description
     * @param orderBy order by
     * @param isAsc is ascending
     * @param offset offset
     * @param limit limit
     * @return list of names
     */
    public List<Name> readNames(Boolean deleted,
            String uuid, String name, String nameEquivalence, String systemStructure, String deviceStructure, String index, String description,
            FieldName orderBy, Boolean isAsc, Integer offset, Integer limit) {
        return readNames(deleted,
                uuid, name, nameEquivalence, systemStructure, deviceStructure, index, description,
                Boolean.FALSE, orderBy, isAsc, offset, limit);
    }

    /**
     * Find names.
     *
     * @param deleted deleted
     * @param uuid uuid
     * @param name name
     * @param nameEquivalence name equivalence
     * @param systemStructure system structure mnemonic
     * @param deviceStructure device structure mnemonic
     * @param index index
     * @param description description
     * @param includeHistory include history
     * @param orderBy order by
     * @param isAsc is ascending
     * @param offset offset
     * @param limit limit
     * @return list of names
     */
    public List<Name> readNames(Boolean deleted,
            String uuid, String name, String nameEquivalence, String systemStructure, String deviceStructure, String index, String description,
            Boolean includeHistory, FieldName orderBy, Boolean isAsc, Integer offset, Integer limit) {

        // note
        //     use of function for mnemonic path
        // where
        //     latest, deleted
        //     queryFields, queryValues
        // order
        //     orderBy, isAsc
        // paging
        //     offset, limit

        CriteriaBuilder cb = em.getCriteriaBuilder();
        CriteriaQuery<Name> cq = cb.createQuery(Name.class);
        Root<Name> from = cq.from(Name.class);

        cq.where(cb.and(preparePredicatesNames(cb, from, deleted,
                uuid, name, nameEquivalence, systemStructure, deviceStructure, index, description,
                includeHistory).toArray(new Predicate[0])));
        cq.select(from);

        if (orderBy != null) {
            if (BooleanUtils.toBoolean(isAsc)) {
                if (FieldName.NAME.equals(orderBy)) {
                    cq.orderBy(cb.asc(from.get(Name.FIELD_CONVENTION_NAME)));
                } else if (FieldName.NAMEEQUIVALENCE.equals(orderBy)) {
                        cq.orderBy(cb.asc(from.get(Name.FIELD_CONVENTION_NAME_EQUIVALENCE)));
                } else if (FieldName.SYSTEMSTRUCTURE.equals(orderBy)) {
                    cq.orderBy(cb.asc(cb.function(NameStructure.FUNCTION_GET_MNEMONIC_PATH_SYSTEM_STRUCTURE, String.class, from.get(Name.FIELD_CONVENTION_NAME))));
                } else if (FieldName.DEVICESTRUCTURE.equals(orderBy)) {
                    cq.orderBy(cb.asc(cb.function(NameStructure.FUNCTION_GET_MNEMONIC_PATH_DEVICE_STRUCTURE, String.class, from.get(Name.FIELD_CONVENTION_NAME))));
                } else if (FieldName.INDEX.equals(orderBy)) {
                    cq.orderBy(cb.asc(cb.function(NameStructure.FUNCTION_GET_INSTANCE_INDEX, String.class, from.get(Name.FIELD_CONVENTION_NAME))));
                } else if (FieldName.DESCRIPTION.equals(orderBy)) {
                    cq.orderBy(cb.asc(from.get(NameStructure.FIELD_DESCRIPTION)));
                } else if (FieldName.WHEN.equals(orderBy)) {
                    cq.orderBy(cb.asc(from.get(NameStructure.FIELD_REQUESTED)));
                } else {
                    cq.orderBy(cb.asc(from.get(Name.FIELD_CONVENTION_NAME)));
                }
            } else {
                if (FieldName.NAME.equals(orderBy)) {
                    cq.orderBy(cb.desc(from.get(Name.FIELD_CONVENTION_NAME)));
                } else if (FieldName.NAMEEQUIVALENCE.equals(orderBy)) {
                    cq.orderBy(cb.desc(from.get(Name.FIELD_CONVENTION_NAME_EQUIVALENCE)));
                } else if (FieldName.SYSTEMSTRUCTURE.equals(orderBy)) {
                    cq.orderBy(cb.desc(cb.function(NameStructure.FUNCTION_GET_MNEMONIC_PATH_SYSTEM_STRUCTURE, String.class, from.get(Name.FIELD_CONVENTION_NAME))));
                } else if (FieldName.DEVICESTRUCTURE.equals(orderBy)) {
                    cq.orderBy(cb.desc(cb.function(NameStructure.FUNCTION_GET_MNEMONIC_PATH_DEVICE_STRUCTURE, String.class, from.get(Name.FIELD_CONVENTION_NAME))));
                } else if (FieldName.INDEX.equals(orderBy)) {
                    cq.orderBy(cb.desc(cb.function(NameStructure.FUNCTION_GET_INSTANCE_INDEX, String.class, from.get(Name.FIELD_CONVENTION_NAME))));
                } else if (FieldName.DESCRIPTION.equals(orderBy)) {
                    cq.orderBy(cb.desc(from.get(NameStructure.FIELD_DESCRIPTION)));
                } else if (FieldName.WHEN.equals(orderBy)) {
                    cq.orderBy(cb.desc(from.get(NameStructure.FIELD_REQUESTED)));
                } else {
                    cq.orderBy(cb.desc(from.get(Name.FIELD_CONVENTION_NAME)));
                }
            }
        }

        TypedQuery<Name> query = em.createQuery(cq);
        if (offset != null && limit != null) {
            query.setFirstResult(offset * limit);
            query.setMaxResults(limit);
        }

        return query.getResultList();
    }

    /**
     * Prepare predicates for names.
     *
     * @param cb criteria builder
     * @param from criteria query root
     * @param deleted deleted
     * @param uuid uuid
     * @param name name
     * @param nameEquivalence name equivalence
     * @param systemStructure system structure mnemonic
     * @param deviceStructure device structure mnemonic
     * @param index index
     * @param description description
     * @param includeHistory include history
     * @return list of predicates
     */
    private List<Predicate> preparePredicatesNames(CriteriaBuilder cb, Root<Name> from, Boolean deleted,
            String uuid, String name, String nameEquivalence, String systemStructure, String deviceStructure, String index, String description,
            Boolean includeHistory) {

        List<Predicate> predicates = new ArrayList<>();

        if (!Boolean.TRUE.equals(includeHistory)) {
            // purpose of Naming to show valid entries
            //     therefore
            //         exclude some values unless history requested
            //         make sure to not exclude present and future values
            //
            // exclude not latest
            //     select * from Name n where
            //         not(n.latest = false)

            Predicate predicateNotLatest = cb.equal(from.get(NameStructure.FIELD_LATEST), Boolean.FALSE);
            Predicate predicateExclude   = cb.not(cb.and(predicateNotLatest));
            predicates.add(predicateExclude);
        }

        if (deleted != null) {
            predicates.add(cb.equal(from.get(NameStructure.FIELD_DELETED), deleted));
        }

        // prepare pattern
        //     jpa query characters % and _
        //     remove excess % characters

        if (!StringUtils.isEmpty(uuid)) {
            predicates.add(cb.and(cb.equal(from.get(NameStructure.FIELD_UUID), RepositoryUtil.preparePattern(uuid))));
        }
        if (!StringUtils.isEmpty(name)) {
            predicates.add(cb.and(cb.like(from.get(Name.FIELD_CONVENTION_NAME), RepositoryUtil.preparePattern(name))));
        }
        if (!StringUtils.isEmpty(nameEquivalence)) {
            predicates.add(cb.and(cb.like(from.get(Name.FIELD_CONVENTION_NAME_EQUIVALENCE), RepositoryUtil.preparePattern(nameEquivalence))));
        }
        if (!StringUtils.isEmpty(systemStructure)) {
            predicates.add(cb.and(cb.like(cb.function(NameStructure.FUNCTION_GET_MNEMONIC_PATH_SYSTEM_STRUCTURE, String.class, from.get(Name.FIELD_CONVENTION_NAME)), RepositoryUtil.preparePattern(systemStructure))));
        }
        if (!StringUtils.isEmpty(deviceStructure)) {
            predicates.add(cb.and(cb.like(cb.function(NameStructure.FUNCTION_GET_MNEMONIC_PATH_DEVICE_STRUCTURE, String.class, from.get(Name.FIELD_CONVENTION_NAME)), RepositoryUtil.preparePattern(deviceStructure))));
        }
        if (!StringUtils.isEmpty(index)) {
            predicates.add(cb.and(cb.like(cb.function(NameStructure.FUNCTION_GET_INSTANCE_INDEX, String.class, from.get(Name.FIELD_CONVENTION_NAME)), RepositoryUtil.preparePattern(index))));
        }
        if (!StringUtils.isEmpty(description)) {
            predicates.add(cb.and(cb.like(from.get(NameStructure.FIELD_DESCRIPTION), RepositoryUtil.preparePattern(description))));
        }

        return predicates;
    }

    /**
     * Persist name into persistence context.
     *
     * @param name name
     */
    public void createName(Name name) {
        em.persist(name);
    }

    /**
     * Merge name into persistence context.
     *
     * @param name name
     */
    public void updateName(Name name) {
        em.merge(name);
    }

}
