/*
 * Copyright (C) 2021 European Spallation Source ERIC.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package org.openepics.names.rest.controller;

import java.io.IOException;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.openepics.names.exception.ServiceException;
import org.openepics.names.rest.api.v1.IStructures;
import org.openepics.names.rest.beans.FieldStructure;
import org.openepics.names.rest.beans.Status;
import org.openepics.names.rest.beans.Type;
import org.openepics.names.rest.beans.element.StructureElement;
import org.openepics.names.rest.beans.element.StructureElementCommand;
import org.openepics.names.rest.beans.element.StructureElementCommandConfirm;
import org.openepics.names.rest.beans.element.StructureElementCommandCreate;
import org.openepics.names.rest.beans.element.StructureElementCommandUpdate;
import org.openepics.names.rest.beans.response.Response;
import org.openepics.names.rest.beans.response.ResponseBoolean;
import org.openepics.names.rest.beans.response.ResponseBooleanList;
import org.openepics.names.rest.beans.response.ResponsePageStructureElements;
import org.openepics.names.service.LogService;
import org.openepics.names.service.StructuresService;
import org.openepics.names.util.ExcelUtil;
import org.openepics.names.util.StructureElementUtil;
import org.openepics.names.util.TextUtil;
import org.openepics.names.util.ValidateStructureElementUtil;
import org.openepics.names.util.ValidateUtil;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.autoconfigure.EnableAutoConfiguration;
import org.springframework.core.io.InputStreamResource;
import org.springframework.core.io.Resource;
import org.springframework.http.HttpHeaders;
import org.springframework.http.HttpStatus;
import org.springframework.http.MediaType;
import org.springframework.http.ResponseEntity;
import org.springframework.web.bind.annotation.RestController;
import org.springframework.web.multipart.MultipartFile;

import com.google.common.collect.Lists;

/**
 * This part of REST API provides structures data for Naming application.
 *
 * @author Lars Johansson
 */
@RestController
@EnableAutoConfiguration
public class StructuresController implements IStructures {

    // TODO validate authority
    //          either
    //              none
    //              naming user
    //              naming admin (& user)

    // note
    //     global exception handler available
    //     controller layer to call validation prior to invoking service layer
    //         business logic may be validation
    //         different kinds of validation

    private static final Logger LOGGER = Logger.getLogger(StructuresController.class.getName());

    private static final String ATTACHMENT_FILENAME_STRUCTURE_ELEMENT_XLSX = "attachment; filename=StructureElement.xlsx";

    private LogService logService;
    private StructuresService structuresService;

    @Autowired
    public StructuresController(
            LogService logUtil,
            StructuresService structuresService) {
        this.logService = logUtil;
        this.structuresService = structuresService;
    }

    @Override
    public ResponseEntity<List<StructureElement>> createStructures(List<StructureElementCommandCreate> structureElementCommands) {
        // validate authority - user & admin
        // convert
        // validate
        // create structures

        try {
            List<StructureElementCommand> commands = StructureElementUtil.convertCommandCreate2Command(structureElementCommands);
            structuresService.validateStructuresCreate(commands);
            return new ResponseEntity<>(structuresService.createStructures(commands, TextUtil.TEST_WHO), Response.getHeaderJson(), HttpStatus.CREATED);
        } catch (ServiceException e) {
            logService.logServiceException(LOGGER, Level.WARNING, e);
            logService.logStackTraceElements(LOGGER, Level.WARNING, e);
            throw e;
        } catch (Exception e) {
            logService.logStackTraceElements(LOGGER, Level.WARNING, e);
            throw e;
        }
    }

    @Override
    public ResponseEntity<Resource> createStructures(MultipartFile file) {
        // validate authority - user & admin
        // convert from input
        // ( convert )
        // validate
        // create structures
        // convert to output

        try {
            if (ExcelUtil.hasExcelFormat(file)) {
                List<StructureElementCommand> structureElementCommands = ExcelUtil.excelToStructureElementCommands(file.getInputStream());
                structuresService.validateStructuresCreate(structureElementCommands);
                List<StructureElement> structureElements = structuresService.createStructures(structureElementCommands, TextUtil.TEST_WHO);
                InputStreamResource isr = new InputStreamResource(ExcelUtil.structureElementsToExcel(structureElements));
                return ResponseEntity.ok()
                        .header(HttpHeaders.CONTENT_DISPOSITION, ATTACHMENT_FILENAME_STRUCTURE_ELEMENT_XLSX)
                        .contentType(MediaType.parseMediaType(ExcelUtil.MIME_TYPE_EXCEL))
                        .body(isr);
            } else {
                throw new ServiceException(TextUtil.FILE_COULD_NOT_BE_PARSED, null, null, null);
            }
       } catch (ServiceException e) {
           logService.logServiceException(LOGGER, Level.WARNING, e);
           logService.logStackTraceElements(LOGGER, Level.WARNING, e);
           throw e;
       } catch (IOException e) {
           logService.logStackTraceElements(LOGGER, Level.WARNING, e);
           throw new ServiceException(TextUtil.FILE_COULD_NOT_BE_PARSED, e.getMessage(), null, e);
       } catch (Exception e) {
           logService.logStackTraceElements(LOGGER, Level.WARNING, e);
           throw e;
       }
    }

    // ----------------------------------------------------------------------------------------------------

    @Override
    public ResponsePageStructureElements readStructures(Type type, Status[] statuses, Boolean deleted,
            String uuid, String parent, String name, String mnemonic, String mnemonicEquivalence, String mnemonicPath, String description,
            FieldStructure orderBy, Boolean isAsc, Integer page, Integer pageSize) {
        // validate
        // read structures

        try {
            ValidateStructureElementUtil.validateStructuresInputRead(type, statuses, deleted,
                    uuid, parent, name, mnemonic, mnemonicEquivalence, mnemonicPath, description,
                    null, orderBy, isAsc, page, pageSize);
            return structuresService.readStructures(type, statuses, deleted,
                    uuid, parent, name, mnemonic, mnemonicEquivalence, mnemonicPath, description,
                    orderBy, isAsc, page, pageSize);
        } catch (ServiceException e) {
            logService.logServiceException(LOGGER, Level.WARNING, e);
            logService.logStackTraceElements(LOGGER, Level.WARNING, e);
            throw e;
        } catch (Exception e) {
            logService.logStackTraceElements(LOGGER, Level.WARNING, e);
            throw e;
        }
    }

    @Override
    public ResponseEntity<Resource> readStructuresDownload(Type type, Status[] statuses, Boolean deleted,
            String uuid, String parent, String name, String mnemonic, String mnemonicEquivalence, String mnemonicPath, String description,
            FieldStructure orderBy, Boolean isAsc, Integer page, Integer pageSize) {
        // validate
        // read structures
        // convert to output

        ResponsePageStructureElements structureElements = readStructures(type, statuses, deleted,
                uuid, parent, name, mnemonic, mnemonicEquivalence, mnemonicPath, description,
                orderBy, isAsc, page, pageSize);
        InputStreamResource isr = new InputStreamResource(ExcelUtil.structureElementsToExcel(structureElements));
        return ResponseEntity.ok()
                .header(HttpHeaders.CONTENT_DISPOSITION, ATTACHMENT_FILENAME_STRUCTURE_ELEMENT_XLSX)
                .contentType(MediaType.parseMediaType(ExcelUtil.MIME_TYPE_EXCEL))
                .body(isr);
    }

    @Override
    public ResponsePageStructureElements readStructuresChildren(String uuid, Type type, Status[] statuses, Boolean deleted,
            String name, String mnemonic, String mnemonicEquivalence, String mnemonicPath, String description,
            FieldStructure orderBy, Boolean isAsc, Integer page, Integer pageSize) {
        // validate
        // read structures

        try {
            ValidateUtil.validateInputUuid(uuid);
            return structuresService.readStructuresChildren(uuid, type, statuses, deleted,
                    name, mnemonic, mnemonicEquivalence, mnemonicPath, description,
                    orderBy, isAsc, page, pageSize);
        } catch (ServiceException e) {
            logService.logServiceException(LOGGER, Level.WARNING, e);
            logService.logStackTraceElements(LOGGER, Level.WARNING, e);
            throw e;
        } catch (Exception e) {
            logService.logStackTraceElements(LOGGER, Level.WARNING, e);
            throw e;
        }
    }

    @Override
    public ResponsePageStructureElements readStructuresMnemonic(String mnemonic,
            FieldStructure orderBy, Boolean isAsc, Integer page, Integer pageSize) {
        // validate
        // read structures

        try {
            ValidateUtil.validateInputMnemonic(mnemonic);
            return structuresService.readStructuresMnemonic(mnemonic, orderBy, isAsc, page, pageSize);
        } catch (ServiceException e) {
            logService.logServiceException(LOGGER, Level.WARNING, e);
            logService.logStackTraceElements(LOGGER, Level.WARNING, e);
            throw e;
        } catch (Exception e) {
            logService.logStackTraceElements(LOGGER, Level.WARNING, e);
            throw e;
        }
    }

    @Override
    public ResponsePageStructureElements readStructuresMnemonicpath(String mnemonicPath,
            FieldStructure orderBy, Boolean isAsc, Integer page, Integer pageSize) {
        // validate
        // read structures

        try {
            ValidateUtil.validateInputMnemonicPath(mnemonicPath);
            return structuresService.readStructuresMnemonicpath(mnemonicPath, orderBy, isAsc, page, pageSize);
        } catch (ServiceException e) {
            logService.logServiceException(LOGGER, Level.WARNING, e);
            logService.logStackTraceElements(LOGGER, Level.WARNING, e);
            throw e;
        } catch (Exception e) {
            logService.logStackTraceElements(LOGGER, Level.WARNING, e);
            throw e;
        }
    }

    @Override
    public ResponsePageStructureElements readStructuresHistory(String uuid, Type type,
            FieldStructure orderBy, Boolean isAsc, Integer page, Integer pageSize) {
        // validate
        // read structures

        try {
            ValidateUtil.validateInputUuid(uuid);
            return structuresService.readStructuresHistory(uuid, type,
                    orderBy, isAsc, page, pageSize);
        } catch (ServiceException e) {
            logService.logServiceException(LOGGER, Level.WARNING, e);
            logService.logStackTraceElements(LOGGER, Level.WARNING, e);
            throw e;
        } catch (Exception e) {
            logService.logStackTraceElements(LOGGER, Level.WARNING, e);
            throw e;
        }
    }

    // ----------------------------------------------------------------------------------------------------

    @Override
    public String equivalenceMnemonic(String mnemonic) {
        // validate
        // read

        try {
            ValidateUtil.validateInputMnemonic(mnemonic);
            return structuresService.equivalenceMnemonic(mnemonic);
        } catch (ServiceException e) {
            logService.logServiceException(LOGGER, Level.WARNING, e);
            logService.logStackTraceElements(LOGGER, Level.WARNING, e);
            throw e;
        } catch (Exception e) {
            logService.logStackTraceElements(LOGGER, Level.WARNING, e);
            throw e;
        }
    }

    @Override
    public ResponseEntity<ResponseBoolean> existsStructure(Type type, String mnemonicPath) {
        // validate
        // read structures

        try {
            ValidateUtil.validateInputType(type);
            ValidateUtil.validateInputMnemonicPath(mnemonicPath);
            return new ResponseEntity<>(new ResponseBoolean(structuresService.existsStructure(type, mnemonicPath)), Response.getHeaderJson(), HttpStatus.OK);
        } catch (ServiceException e) {
            logService.logServiceException(LOGGER, Level.WARNING, e);
            logService.logStackTraceElements(LOGGER, Level.WARNING, e);
            return new ResponseEntity<>(new ResponseBoolean(Boolean.FALSE, e.getMessage(), e.getDetails(), e.getField()), Response.getHeaderJson(), HttpStatus.OK);
        } catch (Exception e) {
            logService.logStackTraceElements(LOGGER, Level.WARNING, e);
            return new ResponseEntity<>(new ResponseBoolean(Boolean.FALSE, TextUtil.OPERATION_COULD_NOT_BE_PERFORMED), Response.getHeaderJson(), HttpStatus.OK);
        }
    }

    @Override
    public ResponseEntity<ResponseBoolean> isValidToCreateStructure(Type type, String mnemonicPath) {
        // validate
        // read structures

        try {
            ValidateUtil.validateInputType(type);
            ValidateUtil.validateInputMnemonicPath(mnemonicPath);
            return new ResponseEntity<>(new ResponseBoolean(structuresService.isValidToCreateStructure(type, mnemonicPath)), Response.getHeaderJson(), HttpStatus.OK);
        } catch (ServiceException e) {
            logService.logServiceException(LOGGER, Level.WARNING, e);
            logService.logStackTraceElements(LOGGER, Level.WARNING, e);
            return new ResponseEntity<>(new ResponseBoolean(Boolean.FALSE, e.getMessage(), e.getDetails(), e.getField()), Response.getHeaderJson(), HttpStatus.OK);
        } catch (Exception e) {
            logService.logStackTraceElements(LOGGER, Level.WARNING, e);
            return new ResponseEntity<>(new ResponseBoolean(Boolean.FALSE, TextUtil.OPERATION_COULD_NOT_BE_PERFORMED), Response.getHeaderJson(), HttpStatus.OK);
        }
    }

    // ----------------------------------------------------------------------------------------------------

    @Override
    public ResponseEntity<ResponseBooleanList> validateStructuresCreate(List<StructureElementCommandCreate> structureElementCommands) {
        // convert
        // validate

        List<StructureElementCommand> commands = StructureElementUtil.convertCommandCreate2Command(structureElementCommands);
        boolean response = true;
        String reason = "";
        List<ResponseBoolean> responses = Lists.newArrayList();
        for (StructureElementCommand command : commands) {
            try {
                structuresService.validateStructuresCreate(command);
                responses.add(new ResponseBoolean(Boolean.TRUE));
            } catch (ServiceException e) {
                logService.logServiceException(LOGGER, Level.WARNING, e);
                logService.logStackTraceElements(LOGGER, Level.WARNING, e);
                if (response) {
                    response = false;
                    reason = TextUtil.ONE_OR_MORE_ELEMENTS_ARE_NOT_CORRECT;
                }
                responses.add(new ResponseBoolean(Boolean.FALSE, e.getMessage(), e.getDetails(), e.getField()));
            } catch (Exception e) {
                logService.logStackTraceElements(LOGGER, Level.WARNING, e);
                if (response) {
                    response = false;
                    reason = TextUtil.ONE_OR_MORE_ELEMENTS_ARE_NOT_CORRECT;
                }
                responses.add(new ResponseBoolean(Boolean.FALSE, TextUtil.OPERATION_COULD_NOT_BE_PERFORMED));
            }
        }
        return new ResponseEntity<>(new ResponseBooleanList(responses, Boolean.valueOf(response), reason), Response.getHeaderJson(), HttpStatus.OK);
    }

    @Override
    public ResponseEntity<ResponseBooleanList> validateStructuresUpdate(List<StructureElementCommandUpdate> structureElementCommands) {
        // convert
        // validate

        List<StructureElementCommand> commands = StructureElementUtil.convertCommandUpdate2Command(structureElementCommands);
        boolean response = true;
        String reason = "";
        List<ResponseBoolean> responses = Lists.newArrayList();
        for (StructureElementCommand command : commands) {
            try {
                structuresService.validateStructuresUpdate(command);
                responses.add(new ResponseBoolean(Boolean.TRUE));
            } catch (ServiceException e) {
                logService.logServiceException(LOGGER, Level.WARNING, e);
                logService.logStackTraceElements(LOGGER, Level.WARNING, e);
                if (response) {
                    response = false;
                    reason = TextUtil.ONE_OR_MORE_ELEMENTS_ARE_NOT_CORRECT;
                }
                responses.add(new ResponseBoolean(Boolean.FALSE, e.getMessage(), e.getDetails(), e.getField()));
            } catch (Exception e) {
                logService.logStackTraceElements(LOGGER, Level.WARNING, e);
                if (response) {
                    response = false;
                    reason = TextUtil.ONE_OR_MORE_ELEMENTS_ARE_NOT_CORRECT;
                }
                responses.add(new ResponseBoolean(Boolean.FALSE, TextUtil.OPERATION_COULD_NOT_BE_PERFORMED));
            }
        }
        return new ResponseEntity<>(new ResponseBooleanList(responses, Boolean.valueOf(response), reason), Response.getHeaderJson(), HttpStatus.OK);
    }

    @Override
    public ResponseEntity<ResponseBooleanList> validateStructuresDelete(List<StructureElementCommandConfirm> structureElementCommands) {
        // convert
        // validate

        List<StructureElementCommand> commands = StructureElementUtil.convertCommandConfirm2Command(structureElementCommands);
        boolean response = true;
        String reason = "";
        List<ResponseBoolean> responses = Lists.newArrayList();
        for (StructureElementCommand command : commands) {
            try {
                structuresService.validateStructuresDelete(command);
                responses.add(new ResponseBoolean(Boolean.TRUE));
            } catch (ServiceException e) {
                logService.logServiceException(LOGGER, Level.WARNING, e);
                logService.logStackTraceElements(LOGGER, Level.WARNING, e);
                if (response) {
                    response = false;
                    reason = TextUtil.ONE_OR_MORE_ELEMENTS_ARE_NOT_CORRECT;
                }
                responses.add(new ResponseBoolean(Boolean.FALSE, e.getMessage(), e.getDetails(), e.getField()));
            } catch (Exception e) {
                logService.logStackTraceElements(LOGGER, Level.WARNING, e);
                if (response) {
                    response = false;
                    reason = TextUtil.ONE_OR_MORE_ELEMENTS_ARE_NOT_CORRECT;
                }
                responses.add(new ResponseBoolean(Boolean.FALSE, TextUtil.OPERATION_COULD_NOT_BE_PERFORMED));
            }
        }
        return new ResponseEntity<>(new ResponseBooleanList(responses, Boolean.valueOf(response), reason), Response.getHeaderJson(), HttpStatus.OK);
    }

    @Override
    public ResponseEntity<ResponseBooleanList> validateStructuresApprove(List<StructureElementCommandConfirm> structureElementCommands) {
        // convert
        // validate

        List<StructureElementCommand> commands = StructureElementUtil.convertCommandConfirm2Command(structureElementCommands);
        boolean response = true;
        String reason = "";
        List<ResponseBoolean> responses = Lists.newArrayList();
        for (StructureElementCommand command : commands) {
            try {
                structuresService.validateStructuresApprove(command);
                responses.add(new ResponseBoolean(Boolean.TRUE));
            } catch (ServiceException e) {
                logService.logServiceException(LOGGER, Level.WARNING, e);
                logService.logStackTraceElements(LOGGER, Level.WARNING, e);
                if (response) {
                    response = false;
                    reason = TextUtil.ONE_OR_MORE_ELEMENTS_ARE_NOT_CORRECT;
                }
                responses.add(new ResponseBoolean(Boolean.FALSE, e.getMessage(), e.getDetails(), e.getField()));
            } catch (Exception e) {
                logService.logStackTraceElements(LOGGER, Level.WARNING, e);
                if (response) {
                    response = false;
                    reason = TextUtil.ONE_OR_MORE_ELEMENTS_ARE_NOT_CORRECT;
                }
                responses.add(new ResponseBoolean(Boolean.FALSE, TextUtil.OPERATION_COULD_NOT_BE_PERFORMED));
            }
        }
        return new ResponseEntity<>(new ResponseBooleanList(responses, Boolean.valueOf(response), reason), Response.getHeaderJson(), HttpStatus.OK);
    }

    @Override
    public ResponseEntity<ResponseBooleanList> validateStructuresCancel(List<StructureElementCommandConfirm> structureElementCommands) {
        // validate

        List<StructureElementCommand> commands = StructureElementUtil.convertCommandConfirm2Command(structureElementCommands);
        boolean response = true;
        String reason = "";
        List<ResponseBoolean> responses = Lists.newArrayList();
        for (StructureElementCommand command : commands) {
            try {
                structuresService.validateStructuresCancel(command);
                responses.add(new ResponseBoolean(Boolean.TRUE));
            } catch (ServiceException e) {
                logService.logServiceException(LOGGER, Level.WARNING, e);
                logService.logStackTraceElements(LOGGER, Level.WARNING, e);
                if (response) {
                    response = false;
                    reason = TextUtil.ONE_OR_MORE_ELEMENTS_ARE_NOT_CORRECT;
                }
                responses.add(new ResponseBoolean(Boolean.FALSE, e.getMessage(), e.getDetails(), e.getField()));
            } catch (Exception e) {
                logService.logStackTraceElements(LOGGER, Level.WARNING, e);
                if (response) {
                    response = false;
                    reason = TextUtil.ONE_OR_MORE_ELEMENTS_ARE_NOT_CORRECT;
                }
                responses.add(new ResponseBoolean(Boolean.FALSE, TextUtil.OPERATION_COULD_NOT_BE_PERFORMED));
            }
        }
        return new ResponseEntity<>(new ResponseBooleanList(responses, Boolean.valueOf(response), reason), Response.getHeaderJson(), HttpStatus.OK);
    }

    @Override
    public ResponseEntity<ResponseBooleanList> validateStructuresReject(List<StructureElementCommandConfirm> structureElementCommands) {
        // convert
        // validate

        List<StructureElementCommand> commands = StructureElementUtil.convertCommandConfirm2Command(structureElementCommands);
        boolean response = true;
        String reason = "";
        List<ResponseBoolean> responses = Lists.newArrayList();
        for (StructureElementCommand command : commands) {
            try {
                structuresService.validateStructuresReject(command);
                responses.add(new ResponseBoolean(Boolean.TRUE));
            } catch (ServiceException e) {
                logService.logServiceException(LOGGER, Level.WARNING, e);
                logService.logStackTraceElements(LOGGER, Level.WARNING, e);
                if (response) {
                    response = false;
                    reason = TextUtil.ONE_OR_MORE_ELEMENTS_ARE_NOT_CORRECT;
                }
                responses.add(new ResponseBoolean(Boolean.FALSE, e.getMessage(), e.getDetails(), e.getField()));
            } catch (Exception e) {
                logService.logStackTraceElements(LOGGER, Level.WARNING, e);
                if (response) {
                    response = false;
                    reason = TextUtil.ONE_OR_MORE_ELEMENTS_ARE_NOT_CORRECT;
                }
                responses.add(new ResponseBoolean(Boolean.FALSE, TextUtil.OPERATION_COULD_NOT_BE_PERFORMED));
            }
        }
        return new ResponseEntity<>(new ResponseBooleanList(responses, Boolean.valueOf(response), reason), Response.getHeaderJson(), HttpStatus.OK);
    }

    // ----------------------------------------------------------------------------------------------------

    @Override
    public List<StructureElement> updateStructures(List<StructureElementCommandUpdate> structureElementCommands) {
        // validate authority - user & admin
        // convert
        // validate
        // update structures

        try {
            List<StructureElementCommand> commands = StructureElementUtil.convertCommandUpdate2Command(structureElementCommands);
            structuresService.validateStructuresUpdate(commands);
            return structuresService.updateStructures(commands, TextUtil.TEST_WHO);
        } catch (ServiceException e) {
            logService.logServiceException(LOGGER, Level.WARNING, e);
            logService.logStackTraceElements(LOGGER, Level.WARNING, e);
            throw e;
        } catch (Exception e) {
            logService.logStackTraceElements(LOGGER, Level.WARNING, e);
            throw e;
        }
    }

    @Override
    public ResponseEntity<Resource> updateStructures(MultipartFile file) {
        // validate authority - user & admin
        // convert from input
        // ( convert )
        // validate
        // update structures
        // convert to output

        try {
            if (ExcelUtil.hasExcelFormat(file)) {
                List<StructureElementCommand> structureElementCommands = ExcelUtil.excelToStructureElementCommands(file.getInputStream());
                structuresService.validateStructuresUpdate(structureElementCommands);
                List<StructureElement> structureElements = structuresService.updateStructures(structureElementCommands, TextUtil.TEST_WHO);
                InputStreamResource isr = new InputStreamResource(ExcelUtil.structureElementsToExcel(structureElements));
                return ResponseEntity.ok()
                        .header(HttpHeaders.CONTENT_DISPOSITION, ATTACHMENT_FILENAME_STRUCTURE_ELEMENT_XLSX)
                        .contentType(MediaType.parseMediaType(ExcelUtil.MIME_TYPE_EXCEL))
                        .body(isr);
            } else {
                throw new ServiceException(TextUtil.FILE_COULD_NOT_BE_PARSED, null, null, null);
            }
       } catch (ServiceException e) {
           logService.logServiceException(LOGGER, Level.WARNING, e);
           logService.logStackTraceElements(LOGGER, Level.WARNING, e);
           throw e;
       } catch (IOException e) {
           logService.logStackTraceElements(LOGGER, Level.WARNING, e);
           throw new ServiceException(TextUtil.FILE_COULD_NOT_BE_PARSED, e.getMessage(), null, e);
       } catch (Exception e) {
           logService.logStackTraceElements(LOGGER, Level.WARNING, e);
           throw e;
       }
    }

    // ----------------------------------------------------------------------------------------------------

    @Override
    public List<StructureElement> deleteStructures(List<StructureElementCommandConfirm> structureElementCommands) {
        // validate authority - user & admin
        // convert
        // validate
        // delete structures

        try {
            List<StructureElementCommand> commands = StructureElementUtil.convertCommandConfirm2Command(structureElementCommands);
            structuresService.validateStructuresDelete(commands);
            return structuresService.deleteStructures(commands, TextUtil.TEST_WHO);
        } catch (ServiceException e) {
            logService.logServiceException(LOGGER, Level.WARNING, e);
            logService.logStackTraceElements(LOGGER, Level.WARNING, e);
            throw e;
        } catch (Exception e) {
            logService.logStackTraceElements(LOGGER, Level.WARNING, e);
            throw e;
        }
    }

    @Override
    public ResponseEntity<Resource> deleteStructures(MultipartFile file) {
        // validate authority - user & admin
        // convert from input
        // ( convert )
        // validate
        // delete structures
        // convert to output

        try {
            if (ExcelUtil.hasExcelFormat(file)) {
                List<StructureElementCommand> structureElementCommands = ExcelUtil.excelToStructureElementCommands(file.getInputStream());
                structuresService.validateStructuresDelete(structureElementCommands);
                List<StructureElement> structureElements = structuresService.deleteStructures(structureElementCommands, TextUtil.TEST_WHO);
                InputStreamResource isr = new InputStreamResource(ExcelUtil.structureElementsToExcel(structureElements));
                return ResponseEntity.ok()
                        .header(HttpHeaders.CONTENT_DISPOSITION, ATTACHMENT_FILENAME_STRUCTURE_ELEMENT_XLSX)
                        .contentType(MediaType.parseMediaType(ExcelUtil.MIME_TYPE_EXCEL))
                        .body(isr);
            } else {
                throw new ServiceException(TextUtil.FILE_COULD_NOT_BE_PARSED, null, null, null);
            }
        } catch (ServiceException e) {
            logService.logServiceException(LOGGER, Level.WARNING, e);
            logService.logStackTraceElements(LOGGER, Level.WARNING, e);
            throw e;
        } catch (IOException e) {
            logService.logStackTraceElements(LOGGER, Level.WARNING, e);
            throw new ServiceException(TextUtil.FILE_COULD_NOT_BE_PARSED, e.getMessage(), null, e);
        } catch (Exception e) {
            logService.logStackTraceElements(LOGGER, Level.WARNING, e);
            throw e;
        }
    }

    // ----------------------------------------------------------------------------------------------------

    @Override
    public List<StructureElement> approveStructures(List<StructureElementCommandConfirm> structureElementCommands) {
        // validate authority - admin
        // convert
        // validate
        // approve structures

        try {
            List<StructureElementCommand> commands = StructureElementUtil.convertCommandConfirm2Command(structureElementCommands);
            structuresService.validateStructuresApprove(commands);
            return structuresService.approveStructures(commands, TextUtil.TEST_WHO);
        } catch (ServiceException e) {
            logService.logServiceException(LOGGER, Level.WARNING, e);
            logService.logStackTraceElements(LOGGER, Level.WARNING, e);
            throw e;
        } catch (Exception e) {
            logService.logStackTraceElements(LOGGER, Level.WARNING, e);
            throw e;
        }
    }

    @Override
    public ResponseEntity<Resource> approveStructures(MultipartFile file) {
        // validate authority - admin
        // convert from input
        // ( convert )
        // validate
        // approve structures
        // convert to output

        try {
            if (ExcelUtil.hasExcelFormat(file)) {
                List<StructureElementCommand> structureElementCommands = ExcelUtil.excelToStructureElementCommands(file.getInputStream());
                structuresService.validateStructuresApprove(structureElementCommands);
                List<StructureElement> structureElements = structuresService.approveStructures(structureElementCommands, TextUtil.TEST_WHO);
                InputStreamResource isr = new InputStreamResource(ExcelUtil.structureElementsToExcel(structureElements));
                return ResponseEntity.ok()
                        .header(HttpHeaders.CONTENT_DISPOSITION, ATTACHMENT_FILENAME_STRUCTURE_ELEMENT_XLSX)
                        .contentType(MediaType.parseMediaType(ExcelUtil.MIME_TYPE_EXCEL))
                        .body(isr);
            } else {
                throw new ServiceException(TextUtil.FILE_COULD_NOT_BE_PARSED, null, null, null);
            }
        } catch (ServiceException e) {
            logService.logServiceException(LOGGER, Level.WARNING, e);
            logService.logStackTraceElements(LOGGER, Level.WARNING, e);
            throw e;
        } catch (IOException e) {
            logService.logStackTraceElements(LOGGER, Level.WARNING, e);
            throw new ServiceException(TextUtil.FILE_COULD_NOT_BE_PARSED, e.getMessage(), null, e);
        } catch (Exception e) {
            logService.logStackTraceElements(LOGGER, Level.WARNING, e);
            throw e;
        }
    }

    @Override
    public List<StructureElement> cancelStructures(List<StructureElementCommandConfirm> structureElementCommands) {
        // validate authority - admin
        // convert
        // validate
        // cancel structures

        try {
            List<StructureElementCommand> commands = StructureElementUtil.convertCommandConfirm2Command(structureElementCommands);
            structuresService.validateStructuresCancel(commands);
            return structuresService.cancelStructures(commands, TextUtil.TEST_WHO);
        } catch (ServiceException e) {
            logService.logServiceException(LOGGER, Level.WARNING, e);
            logService.logStackTraceElements(LOGGER, Level.WARNING, e);
            throw e;
        } catch (Exception e) {
            logService.logStackTraceElements(LOGGER, Level.WARNING, e);
            throw e;
        }
    }

    @Override
    public ResponseEntity<Resource> cancelStructures(MultipartFile file) {
        // validate authority - admin
        // convert from input
        // ( convert )
        // validate
        // cancel structures
        // convert to output

        try {
            if (ExcelUtil.hasExcelFormat(file)) {
                List<StructureElementCommand> structureElementCommands = ExcelUtil.excelToStructureElementCommands(file.getInputStream());
                structuresService.validateStructuresCancel(structureElementCommands);
                List<StructureElement> structureElements = structuresService.cancelStructures(structureElementCommands, TextUtil.TEST_WHO);
                InputStreamResource isr = new InputStreamResource(ExcelUtil.structureElementsToExcel(structureElements));
                return ResponseEntity.ok()
                        .header(HttpHeaders.CONTENT_DISPOSITION, ATTACHMENT_FILENAME_STRUCTURE_ELEMENT_XLSX)
                        .contentType(MediaType.parseMediaType(ExcelUtil.MIME_TYPE_EXCEL))
                        .body(isr);
            } else {
                throw new ServiceException(TextUtil.FILE_COULD_NOT_BE_PARSED, null, null, null);
            }
        } catch (ServiceException e) {
            logService.logServiceException(LOGGER, Level.WARNING, e);
            logService.logStackTraceElements(LOGGER, Level.WARNING, e);
            throw e;
        } catch (IOException e) {
            logService.logStackTraceElements(LOGGER, Level.WARNING, e);
            throw new ServiceException(TextUtil.FILE_COULD_NOT_BE_PARSED, e.getMessage(), null, e);
        } catch (Exception e) {
            logService.logStackTraceElements(LOGGER, Level.WARNING, e);
            throw e;
        }
    }

    @Override
    public List<StructureElement> rejectStructures(List<StructureElementCommandConfirm> structureElementCommands) {
        // validate authority - admin
        // convert
        // validate
        // reject structures

        try {
            List<StructureElementCommand> commands = StructureElementUtil.convertCommandConfirm2Command(structureElementCommands);
            structuresService.validateStructuresReject(commands);
            return structuresService.rejectStructures(commands, TextUtil.TEST_WHO);
        } catch (ServiceException e) {
            logService.logServiceException(LOGGER, Level.WARNING, e);
            logService.logStackTraceElements(LOGGER, Level.WARNING, e);
            throw e;
        } catch (Exception e) {
            logService.logStackTraceElements(LOGGER, Level.WARNING, e);
            throw e;
        }
    }

    @Override
    public ResponseEntity<Resource> rejectStructures(MultipartFile file) {
        // validate authority - admin
        // convert from input
        // ( convert )
        // validate
        // reject structures
        // convert to output

        try {
            if (ExcelUtil.hasExcelFormat(file)) {
                List<StructureElementCommand> structureElementCommands = ExcelUtil.excelToStructureElementCommands(file.getInputStream());
                structuresService.validateStructuresReject(structureElementCommands);
                List<StructureElement> structureElements = structuresService.rejectStructures(structureElementCommands, TextUtil.TEST_WHO);
                InputStreamResource isr = new InputStreamResource(ExcelUtil.structureElementsToExcel(structureElements));
                return ResponseEntity.ok()
                        .header(HttpHeaders.CONTENT_DISPOSITION, ATTACHMENT_FILENAME_STRUCTURE_ELEMENT_XLSX)
                        .contentType(MediaType.parseMediaType(ExcelUtil.MIME_TYPE_EXCEL))
                        .body(isr);
            } else {
                throw new ServiceException(TextUtil.FILE_COULD_NOT_BE_PARSED, null, null, null);
            }
        } catch (ServiceException e) {
            logService.logServiceException(LOGGER, Level.WARNING, e);
            logService.logStackTraceElements(LOGGER, Level.WARNING, e);
            throw e;
        } catch (IOException e) {
            logService.logStackTraceElements(LOGGER, Level.WARNING, e);
            throw new ServiceException(TextUtil.FILE_COULD_NOT_BE_PARSED, e.getMessage(), null, e);
        } catch (Exception e) {
            logService.logStackTraceElements(LOGGER, Level.WARNING, e);
            throw e;
        }
    }

}
