/*
 * Copyright (C) 2021 European Spallation Source ERIC.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package org.openepics.names.service;

import java.text.MessageFormat;
import java.util.Date;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.apache.commons.lang3.StringUtils;
import org.openepics.names.repository.DeviceGroupRepository;
import org.openepics.names.repository.DeviceTypeRepository;
import org.openepics.names.repository.DisciplineRepository;
import org.openepics.names.repository.IDeviceGroupRepository;
import org.openepics.names.repository.IDeviceTypeRepository;
import org.openepics.names.repository.IDisciplineRepository;
import org.openepics.names.repository.INameRepository;
import org.openepics.names.repository.ISubsystemRepository;
import org.openepics.names.repository.ISystemGroupRepository;
import org.openepics.names.repository.ISystemRepository;
import org.openepics.names.repository.NameRepository;
import org.openepics.names.repository.SubsystemRepository;
import org.openepics.names.repository.SystemGroupRepository;
import org.openepics.names.repository.SystemRepository;
import org.openepics.names.repository.model.DeviceGroup;
import org.openepics.names.repository.model.DeviceType;
import org.openepics.names.repository.model.Discipline;
import org.openepics.names.repository.model.Subsystem;
import org.openepics.names.repository.model.System;
import org.openepics.names.repository.model.SystemGroup;
import org.openepics.names.rest.beans.FieldStructure;
import org.openepics.names.rest.beans.Status;
import org.openepics.names.rest.beans.Type;
import org.openepics.names.rest.beans.element.StructureElement;
import org.openepics.names.rest.beans.element.StructureElementCommand;
import org.openepics.names.rest.beans.response.ResponsePageStructureElements;
import org.openepics.names.util.EssNamingConvention;
import org.openepics.names.util.HolderIRepositories;
import org.openepics.names.util.HolderRepositories;
import org.openepics.names.util.HolderSystemDeviceStructure;
import org.openepics.names.util.StructureChoice;
import org.openepics.names.util.StructureCommand;
import org.openepics.names.util.StructureElementUtil;
import org.openepics.names.util.TextUtil;
import org.openepics.names.util.Utilities;
import org.openepics.names.util.ValidateStructureElementUtil;
import org.openepics.names.util.ValidateUtil;
import org.openepics.names.util.notification.NotificationStructure;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import com.google.common.collect.Lists;

/**
 * This class provides structures services.
 *
 * @author Lars Johansson
 */
@Service
public class StructuresService {

    // note
    //     holder
    //         HolderIRepositories and HolderSystemDeviceStructure may or may not be used for preparation of what to return
    //     latest
    //         automatically not show structures that do not come into play
    //         = automatically exclude (approved and not latest)
    //         otherwise refer to history
    //     structurecommand
    //         cud - create update delete
    //         acr - approve cancel reject

    private static final Logger LOGGER = Logger.getLogger(StructuresService.class.getName());

    protected static final String SYSTEM_STRUCTURE_ONLY                 = "System structure only";
    protected static final String DELETE_AFTER_APPROVE_STRUCTURE_CHANGE = "Delete after APPROVE structure change";

    private HolderIRepositories holderIRepositories;
    private HolderRepositories holderRepositories;
    private SystemGroupService systemGroupService;
    private SystemService systemService;
    private SubsystemService subsystemService;
    private DisciplineService disciplineService;
    private DeviceGroupService deviceGroupService;
    private DeviceTypeService deviceTypeService;
    private NotificationService notificationService;
    private EssNamingConvention namingConvention;

    @Autowired
    public StructuresService(
            INameRepository iNameRepository,
            ISystemGroupRepository iSystemGroupRepository,
            ISystemRepository iSystemRepository,
            ISubsystemRepository iSubsystemRepository,
            IDisciplineRepository iDisciplineRepository,
            IDeviceGroupRepository iDeviceGroupRepository,
            IDeviceTypeRepository iDeviceTypeRepository,
            NameRepository nameRepository,
            SystemGroupRepository systemGroupRepository,
            SystemRepository systemRepository,
            SubsystemRepository subsystemRepository,
            DisciplineRepository disciplineRepository,
            DeviceGroupRepository deviceGroupRepository,
            DeviceTypeRepository deviceTypeRepository,
            SystemGroupService systemGroupService,
            SystemService systemService,
            SubsystemService subsystemService,
            DisciplineService disciplineService,
            DeviceGroupService deviceGroupService,
            DeviceTypeService deviceTypeService,
            NotificationService notificationService) {

        this.namingConvention = new EssNamingConvention();
        this.holderIRepositories = new HolderIRepositories(
                iNameRepository,
                iSystemGroupRepository,
                iSystemRepository,
                iSubsystemRepository,
                iDisciplineRepository,
                iDeviceGroupRepository,
                iDeviceTypeRepository);
        this.holderRepositories = new HolderRepositories(
                nameRepository,
                systemGroupRepository,
                systemRepository,
                subsystemRepository,
                disciplineRepository,
                deviceGroupRepository,
                deviceTypeRepository);
        this.systemGroupService = systemGroupService;
        this.systemService = systemService;
        this.subsystemService = subsystemService;
        this.disciplineService = disciplineService;
        this.deviceGroupService = deviceGroupService;
        this.deviceTypeService = deviceTypeService;
        this.notificationService = notificationService;
    }

    @Transactional
    public List<StructureElement> createStructures(List<StructureElementCommand> structureElements, String username) {
        // validation outside method
        // transaction
        //     for each structure element command
        //         create structure
        //         handle
        //             notification
        //             structure element for created structure
        //     notify
        //     return structure elements for created structures

        // initiate holder of system and device structure content, for performance reasons
        HolderSystemDeviceStructure holderStructure = new HolderSystemDeviceStructure(holderIRepositories);

        Date when = new Date();
        List<NotificationStructure> notifications = Lists.newArrayList();
        final List<StructureElement> createdStructureElements = Lists.newArrayList();
        for (StructureElementCommand structureElement : structureElements) {
            Type type = structureElement.getType();
            StructureElementNotification structureElementNotification = null;

            if (Type.SYSTEMGROUP.equals(type)) {
                structureElementNotification = systemGroupService.createStructure(structureElement, when, username, namingConvention, holderStructure);
            } else if (Type.SYSTEM.equals(type)) {
                structureElementNotification = systemService.createStructure(structureElement, when, username, namingConvention, holderStructure);
            } else if (Type.SUBSYSTEM.equals(type)) {
                structureElementNotification = subsystemService.createStructure(structureElement, when, username, namingConvention, holderStructure);
            } else if (Type.DISCIPLINE.equals(type)) {
                structureElementNotification = disciplineService.createStructure(structureElement, when, username, namingConvention, holderStructure);
            } else if (Type.DEVICEGROUP.equals(type)) {
                structureElementNotification = deviceGroupService.createStructure(structureElement, when, username, namingConvention, holderStructure);
            } else if (Type.DEVICETYPE.equals(type)) {
                structureElementNotification = deviceTypeService.createStructure(structureElement, when, username, namingConvention, holderStructure);
            } else {
                continue;
            }

            Utilities.addToCollection(notifications, structureElementNotification.notificationStructure());
            Utilities.addToCollection(createdStructureElements, structureElementNotification.structureElement());

            if (LOGGER.isLoggable(Level.FINER)) {
                LOGGER.log(Level.FINER, MessageFormat.format(TextUtil.DESCRIPTION_NAME_VALUE, TextUtil.CREATE_STRUCTURE, "type", type));
                LOGGER.log(Level.FINER, MessageFormat.format(TextUtil.DESCRIPTION_NAME_VALUE, TextUtil.CREATE_STRUCTURE, TextUtil.ELEMENT_IN, structureElement));
                LOGGER.log(Level.FINER, MessageFormat.format(TextUtil.DESCRIPTION_NAME_VALUE, TextUtil.CREATE_STRUCTURE, TextUtil.ELEMENT_OUT, structureElementNotification.structureElement()));
                LOGGER.log(Level.FINER, MessageFormat.format(TextUtil.DESCRIPTION_NAME_VALUE, TextUtil.CREATE_STRUCTURE, TextUtil.NOTIFICATION, structureElementNotification.notificationStructure()));
            }
        }

        // notify
        notificationService.notifyStructures(notifications, StructureCommand.CREATE);

        LOGGER.log(Level.INFO,
                () -> MessageFormat.format(
                        TextUtil.DESCRIPTION_NUMBER_ELEMENTS_IN_OUT,
                        "Create structures",
                        structureElements.size(),
                        createdStructureElements.size()));
        return createdStructureElements;
    }

    // ----------------------------------------------------------------------------------------------------

    public ResponsePageStructureElements readStructures(Type type, Status[] statuses, Boolean deleted,
            String uuid, String parent, String name, String mnemonic, String mnemonicEquivalence, String mnemonicPath, String description,
            FieldStructure orderBy, Boolean isAsc, Integer offset, Integer limit) {
        return readStructures(type, statuses, deleted,
                uuid, parent, name, mnemonic, mnemonicEquivalence, mnemonicPath, description,
                Boolean.FALSE, orderBy, isAsc, offset, limit, StructureChoice.STRUCTURE);
    }

    public ResponsePageStructureElements readStructures(Type type, Status[] statuses, Boolean deleted,
            String uuid, String parent, String name, String mnemonic, String mnemonicEquivalence, String mnemonicPath, String description,
            Boolean includeHistory, FieldStructure orderBy, Boolean isAsc, Integer offset, Integer limit,
            StructureChoice structureChoice) {
        // validation outside method
        // read structures
        // return structure elements for structures

        if (LOGGER.isLoggable(Level.FINE)) {
            LOGGER.log(Level.FINE, MessageFormat.format(TextUtil.DESCRIPTION_NAME_VALUE, TextUtil.READ_STRUCTURES, "type", type));
            LOGGER.log(Level.FINE, MessageFormat.format(TextUtil.DESCRIPTION_NAME_VALUE, TextUtil.READ_STRUCTURES, "statuses.length", Utilities.getLength(statuses)));
            LOGGER.log(Level.FINE, MessageFormat.format(TextUtil.DESCRIPTION_NAME_VALUE, TextUtil.READ_STRUCTURES, "deleted", deleted));
            LOGGER.log(Level.FINE, MessageFormat.format(TextUtil.DESCRIPTION_NAME_VALUE, TextUtil.READ_STRUCTURES, "uuid", uuid));
            LOGGER.log(Level.FINE, MessageFormat.format(TextUtil.DESCRIPTION_NAME_VALUE, TextUtil.READ_STRUCTURES, "parent", parent));
            LOGGER.log(Level.FINE, MessageFormat.format(TextUtil.DESCRIPTION_NAME_VALUE, TextUtil.READ_STRUCTURES, "name", name));
            LOGGER.log(Level.FINE, MessageFormat.format(TextUtil.DESCRIPTION_NAME_VALUE, TextUtil.READ_STRUCTURES, "mnemonic", mnemonic));
            LOGGER.log(Level.FINE, MessageFormat.format(TextUtil.DESCRIPTION_NAME_VALUE, TextUtil.READ_STRUCTURES, "mnemonicEquivalence", mnemonicEquivalence));
            LOGGER.log(Level.FINE, MessageFormat.format(TextUtil.DESCRIPTION_NAME_VALUE, TextUtil.READ_STRUCTURES, "mnemonicPath", mnemonicPath));
            LOGGER.log(Level.FINE, MessageFormat.format(TextUtil.DESCRIPTION_NAME_VALUE, TextUtil.READ_STRUCTURES, "description", description));
            LOGGER.log(Level.FINE, MessageFormat.format(TextUtil.DESCRIPTION_NAME_VALUE, TextUtil.READ_STRUCTURES, "includeHistory", includeHistory));
            LOGGER.log(Level.FINE, MessageFormat.format(TextUtil.DESCRIPTION_NAME_VALUE, TextUtil.READ_STRUCTURES, "orderBy", orderBy));
            LOGGER.log(Level.FINE, MessageFormat.format(TextUtil.DESCRIPTION_NAME_VALUE, TextUtil.READ_STRUCTURES, "isAsc", isAsc));
            LOGGER.log(Level.FINE, MessageFormat.format(TextUtil.DESCRIPTION_NAME_VALUE, TextUtil.READ_STRUCTURES, "offset", offset));
            LOGGER.log(Level.FINE, MessageFormat.format(TextUtil.DESCRIPTION_NAME_VALUE, TextUtil.READ_STRUCTURES, "limit", limit));
            LOGGER.log(Level.FINE, MessageFormat.format(TextUtil.DESCRIPTION_NAME_VALUE, TextUtil.READ_STRUCTURES, "structureChoice", structureChoice));
            if (statuses != null) {
                for (Status status : statuses) {
                    LOGGER.log(Level.FINE, MessageFormat.format(TextUtil.DESCRIPTION_NAME_VALUE, TextUtil.READ_STRUCTURES, "status", status));
                }
            }
        }

        // initiate holder of system and device structure content, for performance reasons
        HolderSystemDeviceStructure holderStructure = new HolderSystemDeviceStructure(holderIRepositories);

        final List<StructureElement> structureElements = Lists.newArrayList();
        Long totalCount = null;
        if (Type.SYSTEMGROUP.equals(type)) {
            // system group has no parent uuid
            List<SystemGroup> systemGroups = null;
            if (StringUtils.isEmpty(parent)) {
                systemGroups = holderRepositories.systemGroupRepository().readSystemGroups(statuses, deleted,
                        uuid, name, mnemonic, mnemonicEquivalence, mnemonicPath, description,
                        includeHistory, orderBy, isAsc, offset, limit);
                totalCount = holderRepositories.systemGroupRepository().countSystemGroups(statuses, deleted,
                        uuid, name, mnemonic, mnemonicEquivalence, mnemonicPath, description,
                        includeHistory);
            } else {
                systemGroups = Lists.newArrayList();
                totalCount = 0L;
            }
            structureElements.addAll(StructureElementUtil.getStructureElementsForSystemGroups(systemGroups, holderStructure, structureChoice));
        } else if (Type.SYSTEM.equals(type)) {
            List<System> systems = holderRepositories.systemRepository().readSystems(statuses, deleted,
                    uuid, parent, name, mnemonic, mnemonicEquivalence, mnemonicPath, description,
                    includeHistory, orderBy, isAsc, offset, limit);
            totalCount = holderRepositories.systemRepository().countSystems(statuses, deleted,
                    uuid, parent, name, mnemonic, mnemonicEquivalence, mnemonicPath, description,
                    includeHistory);
            structureElements.addAll(StructureElementUtil.getStructureElementsForSystems(systems, holderStructure, structureChoice));
        } else if (Type.SUBSYSTEM.equals(type)) {
            List<Subsystem> subsystems = holderRepositories.subsystemRepository().readSubsystems(statuses, deleted,
                    uuid, parent, name, mnemonic, mnemonicEquivalence, mnemonicPath, description,
                    includeHistory, orderBy, isAsc, offset, limit);
            totalCount = holderRepositories.subsystemRepository().countSubsystems(statuses, deleted,
                    uuid, parent, name, mnemonic, mnemonicEquivalence, mnemonicPath, description,
                    includeHistory);
            structureElements.addAll(StructureElementUtil.getStructureElementsForSubsystems(subsystems, holderStructure, structureChoice));
        } else if (Type.DISCIPLINE.equals(type)) {
            // discipline has no parent uuid
            List<Discipline> disciplines = null;
            if (StringUtils.isEmpty(parent)) {
                disciplines = holderRepositories.disciplineRepository().readDisciplines(statuses, deleted,
                        uuid, name, mnemonic, mnemonicEquivalence, mnemonicPath, description,
                        includeHistory, orderBy, isAsc, offset, limit);
                totalCount = holderRepositories.disciplineRepository().countDisciplines(statuses, deleted,
                        uuid, name, mnemonic, mnemonicEquivalence, mnemonicPath, description,
                        includeHistory);
            } else {
                disciplines = Lists.newArrayList();
                totalCount = 0L;
            }
            structureElements.addAll(StructureElementUtil.getStructureElementsForDisciplines(disciplines, holderStructure, structureChoice));
        } else if (Type.DEVICEGROUP.equals(type)) {
            List<DeviceGroup> deviceGroups = holderRepositories.deviceGroupRepository().readDeviceGroups(statuses, deleted,
                    uuid, parent, name, mnemonic, mnemonicEquivalence, mnemonicPath, description,
                    includeHistory, orderBy, isAsc, offset, limit);
            totalCount = holderRepositories.deviceGroupRepository().countDeviceGroups(statuses, deleted,
                    uuid, parent, name, mnemonic, mnemonicEquivalence, mnemonicPath, description,
                    includeHistory);
            structureElements.addAll(StructureElementUtil.getStructureElementsForDeviceGroups(deviceGroups, holderStructure, structureChoice));
        } else if (Type.DEVICETYPE.equals(type)) {
            List<DeviceType> deviceTypes = holderRepositories.deviceTypeRepository().readDeviceTypes(statuses, deleted,
                    uuid, parent, name, mnemonic, mnemonicEquivalence, mnemonicPath, description,
                    includeHistory, orderBy, isAsc, offset, limit);
            totalCount = holderRepositories.deviceTypeRepository().countDeviceTypes(statuses, deleted,
                    uuid, parent, name, mnemonic, mnemonicEquivalence, mnemonicPath, description,
                    includeHistory);
            structureElements.addAll(StructureElementUtil.getStructureElementsForDeviceTypes(deviceTypes, holderStructure, structureChoice));
        }

        ResponsePageStructureElements response = new ResponsePageStructureElements(structureElements, totalCount, structureElements.size(), offset, limit);
        LOGGER.log(Level.FINE,
                () -> MessageFormat.format(
                        TextUtil.DESCRIPTION_NUMBER_ELEMENTS,
                        TextUtil.READ_STRUCTURES,
                        structureElements.size()));
        return response;
    }

    public ResponsePageStructureElements readStructuresChildren(String uuid, Type type,
            Status[] statuses, Boolean deleted, String name, String mnemonic, String mnemonicEquivalence, String mnemonicPath, String description,
            FieldStructure orderBy, Boolean isAsc, Integer offset, Integer limit) {
        // validation outside method
        // read structure latest by uuid for type
        // return structure elements for structures

        if (LOGGER.isLoggable(Level.FINE)) {
            LOGGER.log(Level.FINE, MessageFormat.format(TextUtil.DESCRIPTION_NAME_VALUE, "Read structures children", "uuid", uuid));
            LOGGER.log(Level.FINE, MessageFormat.format(TextUtil.DESCRIPTION_NAME_VALUE, "Read structures children", "type", type));
        }

        boolean typeSystemGroup = Type.SYSTEMGROUP.equals(type);
        boolean typeSystem      = Type.SYSTEM.equals(type);
        boolean typeSubsystem   = Type.SUBSYSTEM.equals(type);
        boolean typeDiscipline  = Type.DISCIPLINE.equals(type);
        boolean typeDeviceGroup = Type.DEVICEGROUP.equals(type);
        boolean typeDeviceType  = Type.DEVICETYPE.equals(type);
        boolean typeEither      = ValidateUtil.isAnyTrue(typeSystemGroup, typeSystem, typeSubsystem, typeDiscipline, typeDeviceGroup, typeDeviceType);

        final List<StructureElement> structureElements = Lists.newArrayList();
        ResponsePageStructureElements response = null;
        if (typeEither) {
            if (typeSystemGroup) {
                response = readStructures(Type.SYSTEM,      new Status[] {Status.APPROVED}, false, null, uuid, null, null, null, null, null, orderBy, isAsc, offset, limit);
            } else if (typeSystem) {
                response = readStructures(Type.SUBSYSTEM,   new Status[] {Status.APPROVED}, false, null, uuid, null, null, null, null, null, orderBy, isAsc, offset, limit);
            } else if (typeDiscipline) {
                response = readStructures(Type.DEVICEGROUP, new Status[] {Status.APPROVED}, false, null, uuid, null, null, null, null, null, orderBy, isAsc, offset, limit);
            } else if (typeDeviceGroup) {
                response = readStructures(Type.DEVICETYPE,  new Status[] {Status.APPROVED}, false, null, uuid, null, null, null, null, null, orderBy, isAsc, offset, limit);
            } else {
                // no children in structures for subsystem and device type
                response = new ResponsePageStructureElements(structureElements, Long.valueOf(structureElements.size()), structureElements.size(), offset, limit);
            }
        } else {
            // go through all structures and see if / where uuid is found
            //     uuid in 0 or 1 of structures

            ResponsePageStructureElements responseSystemGroup = readStructures(Type.SYSTEM,      new Status[] {Status.APPROVED}, false, null, uuid, null, null, null, null, null, orderBy, isAsc, offset, limit);
            ResponsePageStructureElements responseSystem      = readStructures(Type.SUBSYSTEM,   new Status[] {Status.APPROVED}, false, null, uuid, null, null, null, null, null, orderBy, isAsc, offset, limit);
            ResponsePageStructureElements responseDiscipline  = readStructures(Type.DEVICEGROUP, new Status[] {Status.APPROVED}, false, null, uuid, null, null, null, null, null, orderBy, isAsc, offset, limit);
            ResponsePageStructureElements responseDeviceGroup = readStructures(Type.DEVICETYPE,  new Status[] {Status.APPROVED}, false, null, uuid, null, null, null, null, null, orderBy, isAsc, offset, limit);

            structureElements.addAll(responseSystemGroup.getList());
            structureElements.addAll(responseSystem.getList());
            structureElements.addAll(responseDiscipline.getList());
            structureElements.addAll(responseDeviceGroup.getList());

            // uuid is in max one structure
            response = new ResponsePageStructureElements(structureElements, Long.valueOf(structureElements.size()), structureElements.size(), offset, limit);
        }
        return response;
    }

    public ResponsePageStructureElements readStructuresMnemonic(String mnemonic,
            FieldStructure orderBy, Boolean isAsc, Integer offset, Integer limit) {
        // validation outside method
        // read structure latest by mnemonic
        // return structure elements for structures

        LOGGER.log(Level.FINE, () -> MessageFormat.format(TextUtil.DESCRIPTION_NAME_VALUE, "Read structures mnemonic", "mnemonic", mnemonic));

        // TODO handle orderBy, isAsc, offset, limit

        ResponsePageStructureElements responseSystemGroup = readStructures(Type.SYSTEMGROUP, new Status[] {Status.APPROVED}, false, null, null, null, mnemonic, null, null, null, orderBy, isAsc, offset, limit);
        ResponsePageStructureElements responseSystem      = readStructures(Type.SYSTEM,      new Status[] {Status.APPROVED}, false, null, null, null, mnemonic, null, null, null, orderBy, isAsc, offset, limit);
        ResponsePageStructureElements responseSubsystem   = readStructures(Type.SUBSYSTEM,   new Status[] {Status.APPROVED}, false, null, null, null, mnemonic, null, null, null, orderBy, isAsc, offset, limit);
        ResponsePageStructureElements responseDiscipline  = readStructures(Type.DISCIPLINE,  new Status[] {Status.APPROVED}, false, null, null, null, mnemonic, null, null, null, orderBy, isAsc, offset, limit);
        ResponsePageStructureElements responseDeviceGroup = readStructures(Type.DEVICEGROUP, new Status[] {Status.APPROVED}, false, null, null, null, mnemonic, null, null, null, orderBy, isAsc, offset, limit);
        ResponsePageStructureElements responseDeviceType  = readStructures(Type.DEVICETYPE,  new Status[] {Status.APPROVED}, false, null, null, null, mnemonic, null, null, null, orderBy, isAsc, offset, limit);

        List<StructureElement> structureElements = Lists.newArrayList();
        structureElements.addAll(responseSystemGroup.getList());
        structureElements.addAll(responseSystem.getList());
        structureElements.addAll(responseSubsystem.getList());
        structureElements.addAll(responseDiscipline.getList());
        structureElements.addAll(responseDeviceGroup.getList());
        structureElements.addAll(responseDeviceType.getList());

        return new ResponsePageStructureElements(structureElements, Long.valueOf(structureElements.size()), structureElements.size(), 0, structureElements.size());
    }

    public ResponsePageStructureElements readStructuresMnemonicpath(String mnemonicPath,
            FieldStructure orderBy, Boolean isAsc, Integer offset, Integer limit) {
        // validation outside method
        // read structure latest by mnemonic path
        // return structure elements for structures

        LOGGER.log(Level.FINE, () -> MessageFormat.format(TextUtil.DESCRIPTION_NAME_VALUE, "Read structures mnemonic path", "mnemonicPath", mnemonicPath));

        // TODO handle orderBy, isAsc, offset, limit

        ResponsePageStructureElements responseSystemGroup = readStructures(Type.SYSTEMGROUP, new Status[] {Status.APPROVED}, false, null, null, null, null, null, mnemonicPath, null, orderBy, isAsc, offset, limit);
        ResponsePageStructureElements responseSystem      = readStructures(Type.SYSTEM,      new Status[] {Status.APPROVED}, false, null, null, null, null, null, mnemonicPath, null, orderBy, isAsc, offset, limit);
        ResponsePageStructureElements responseSubsystem   = readStructures(Type.SUBSYSTEM,   new Status[] {Status.APPROVED}, false, null, null, null, null, null, mnemonicPath, null, orderBy, isAsc, offset, limit);
        ResponsePageStructureElements responseDiscipline  = readStructures(Type.DISCIPLINE,  new Status[] {Status.APPROVED}, false, null, null, null, null, null, mnemonicPath, null, orderBy, isAsc, offset, limit);
        ResponsePageStructureElements responseDeviceGroup = readStructures(Type.DEVICEGROUP, new Status[] {Status.APPROVED}, false, null, null, null, null, null, mnemonicPath, null, orderBy, isAsc, offset, limit);
        ResponsePageStructureElements responseDeviceType  = readStructures(Type.DEVICETYPE,  new Status[] {Status.APPROVED}, false, null, null, null, null, null, mnemonicPath, null, orderBy, isAsc, offset, limit);

        List<StructureElement> structureElements = Lists.newArrayList();
        structureElements.addAll(responseSystemGroup.getList());
        structureElements.addAll(responseSystem.getList());
        structureElements.addAll(responseSubsystem.getList());
        structureElements.addAll(responseDiscipline.getList());
        structureElements.addAll(responseDeviceGroup.getList());
        structureElements.addAll(responseDeviceType.getList());

        return new ResponsePageStructureElements(structureElements, Long.valueOf(structureElements.size()), structureElements.size(), 0, structureElements.size());
    }

    /**
     * Read history for structure by uuid.
     *
     * @param uuid uuid
     * @param type type
     * @return list of structure elements
     */
    public ResponsePageStructureElements readStructuresHistory(String uuid, Type type,
            FieldStructure orderBy, Boolean isAsc, Integer offset, Integer limit) {
        // validation outside method
        // read history for structure
        // return structure elements for structures
        // note
        //     type may speed up read

        if (LOGGER.isLoggable(Level.FINE)) {
            LOGGER.log(Level.FINE, MessageFormat.format(TextUtil.DESCRIPTION_NAME_VALUE, "Read structures history", "uuid", uuid));
            LOGGER.log(Level.FINE, MessageFormat.format(TextUtil.DESCRIPTION_NAME_VALUE, "Read structures history", "type", type));
        }

        boolean typeSystemGroup = Type.SYSTEMGROUP.equals(type);
        boolean typeSystem      = Type.SYSTEM.equals(type);
        boolean typeSubsystem   = Type.SUBSYSTEM.equals(type);
        boolean typeDiscipline  = Type.DISCIPLINE.equals(type);
        boolean typeDeviceGroup = Type.DEVICEGROUP.equals(type);
        boolean typeDeviceType  = Type.DEVICETYPE.equals(type);
        boolean typeEither      = ValidateUtil.isAnyTrue(typeSystemGroup, typeSystem, typeSubsystem, typeDiscipline, typeDeviceGroup, typeDeviceType);

        // mnemonic path does not make same sense for history
        // (very) tricky to find mnemonic path for uuid at proper time (history)
        // therefore empty mnemonic path for history for structure

        final List<StructureElement> structureElements = Lists.newArrayList();
        ResponsePageStructureElements response = null;
        if (typeEither) {
            if (typeSystemGroup) {
                response = readStructures(Type.SYSTEMGROUP, null, null, uuid, null, null, null, null, null, null, Boolean.TRUE, orderBy, isAsc, offset, limit, StructureChoice.HISTORY);
            } else if (typeSystem) {
                response = readStructures(Type.SYSTEM,      null, null, uuid, null, null, null, null, null, null, Boolean.TRUE, orderBy, isAsc, offset, limit, StructureChoice.HISTORY);
            } else if (typeSubsystem) {
                response = readStructures(Type.SUBSYSTEM,   null, null, uuid, null, null, null, null, null, null, Boolean.TRUE, orderBy, isAsc, offset, limit, StructureChoice.HISTORY);
            } else if (typeDiscipline) {
                response = readStructures(Type.DISCIPLINE,  null, null, uuid, null, null, null, null, null, null, Boolean.TRUE, orderBy, isAsc, offset, limit, StructureChoice.HISTORY);
            } else if (typeDeviceGroup) {
                response = readStructures(Type.DEVICEGROUP, null, null, uuid, null, null, null, null, null, null, Boolean.TRUE, orderBy, isAsc, offset, limit, StructureChoice.HISTORY);
            } else if (typeDeviceType) {
                response = readStructures(Type.DEVICETYPE,  null, null, uuid, null, null, null, null, null, null, Boolean.TRUE, orderBy, isAsc, offset, limit, StructureChoice.HISTORY);
            }
        } else {
            // go through all structures and see if / where uuid is found
            //     uuid in 0 or 1 of structures

            ResponsePageStructureElements responseSystemGroup = readStructures(Type.SYSTEMGROUP, null, null, uuid, null, null, null, null, null, null, Boolean.TRUE, orderBy, isAsc, offset, limit, StructureChoice.HISTORY);
            ResponsePageStructureElements responseSystem      = readStructures(Type.SYSTEM,      null, null, uuid, null, null, null, null, null, null, Boolean.TRUE, orderBy, isAsc, offset, limit, StructureChoice.HISTORY);
            ResponsePageStructureElements responseSubsystem   = readStructures(Type.SUBSYSTEM,   null, null, uuid, null, null, null, null, null, null, Boolean.TRUE, orderBy, isAsc, offset, limit, StructureChoice.HISTORY);
            ResponsePageStructureElements responseDiscipline  = readStructures(Type.DISCIPLINE,  null, null, uuid, null, null, null, null, null, null, Boolean.TRUE, orderBy, isAsc, offset, limit, StructureChoice.HISTORY);
            ResponsePageStructureElements responseDeviceGroup = readStructures(Type.DEVICEGROUP, null, null, uuid, null, null, null, null, null, null, Boolean.TRUE, orderBy, isAsc, offset, limit, StructureChoice.HISTORY);
            ResponsePageStructureElements responseDeviceType  = readStructures(Type.DEVICETYPE,  null, null, uuid, null, null, null, null, null, null, Boolean.TRUE, orderBy, isAsc, offset, limit, StructureChoice.HISTORY);

            structureElements.addAll(responseSystemGroup.getList());
            structureElements.addAll(responseSystem.getList());
            structureElements.addAll(responseSubsystem.getList());
            structureElements.addAll(responseDiscipline.getList());
            structureElements.addAll(responseDeviceGroup.getList());
            structureElements.addAll(responseDeviceType.getList());

            // uuid is in max one structure
            response = new ResponsePageStructureElements(structureElements, Long.valueOf(structureElements.size()), structureElements.size(), offset, limit);

        }

        return response;
    }

    // ----------------------------------------------------------------------------------------------------

    public String equivalenceMnemonic(String mnemonic) {
        // validation outside method
        // read equivalence

        LOGGER.log(Level.FINE, () -> MessageFormat.format(TextUtil.DESCRIPTION_NAME_VALUE, "Equivalence mnemonic", "mnemonic", mnemonic));

        return namingConvention.equivalenceClassRepresentative(mnemonic);
    }

    public Boolean existsStructure(Type type, String mnemonicPath) {
        // validation outside method
        // read exists

        if (LOGGER.isLoggable(Level.FINE)) {
            LOGGER.log(Level.FINE, MessageFormat.format(TextUtil.DESCRIPTION_NAME_VALUE, "Exists structure", "type", type));
            LOGGER.log(Level.FINE, MessageFormat.format(TextUtil.DESCRIPTION_NAME_VALUE, "Exists structure", "mnemonicPath", mnemonicPath));
        }

        if (Type.SYSTEMGROUP.equals(type)) {
            List<SystemGroup> systemGroups = holderRepositories.systemGroupRepository().readSystemGroups(Status.APPROVED, false, null, null, null, null, mnemonicPath, null);
            return !systemGroups.isEmpty();
        } else if (Type.SYSTEM.equals(type)) {
            List<System> systems = holderRepositories.systemRepository().readSystems(Status.APPROVED, false, null, null, null, null, null, mnemonicPath, null);
            return !systems.isEmpty();
        } else if (Type.SUBSYSTEM.equals(type)) {
            List<Subsystem> subsystems = holderRepositories.subsystemRepository().readSubsystems(Status.APPROVED, false, null, null, null, null, null, mnemonicPath, null);
            return !subsystems.isEmpty();
        } else if (Type.DISCIPLINE.equals(type)) {
            List<Discipline> disciplines = holderRepositories.disciplineRepository().readDisciplines(Status.APPROVED, false, null, null, null, null, mnemonicPath, null);
            return !disciplines.isEmpty();
        } else if (Type.DEVICEGROUP.equals(type)) {
            return Boolean.FALSE;
        } else if (Type.DEVICETYPE.equals(type)) {
            List<DeviceType> deviceTypes = holderRepositories.deviceTypeRepository().readDeviceTypes(Status.APPROVED, false, null, null, null, null, null, mnemonicPath, null);
            return !deviceTypes.isEmpty();
        }

        return Boolean.FALSE;
    }

    public Boolean isValidToCreateStructure(Type type, String mnemonicPath) {
        // validation outside method
        // validate data - not exists

        if (LOGGER.isLoggable(Level.FINE)) {
            LOGGER.log(Level.FINE, MessageFormat.format(TextUtil.DESCRIPTION_NAME_VALUE, "Is valid to create structure", "type", type));
            LOGGER.log(Level.FINE, MessageFormat.format(TextUtil.DESCRIPTION_NAME_VALUE, "Is valid to create structure", "mnemonicPath", mnemonicPath));
        }

        // initiate holder of system and device structure content, for performance reasons
        //     note false to not include deleted entries
        HolderSystemDeviceStructure holderStructure = new HolderSystemDeviceStructure(holderIRepositories, false);

        ValidateStructureElementUtil.validateStructureDataCreate(type, mnemonicPath, namingConvention, holderIRepositories, holderStructure);

        return Boolean.TRUE;
    }

    // ----------------------------------------------------------------------------------------------------

    public void validateStructuresCreate(StructureElementCommand structureElement) {
        validateStructuresCreate(structureElement, new HolderSystemDeviceStructure(holderIRepositories));
    }
    public void validateStructuresCreate(StructureElementCommand structureElement, HolderSystemDeviceStructure holderStructure) {
        // validate structure element
        //     input
        //         type
        //         input itself
        //     data
        //         data itself
        //         relative other data

        // validate input
        ValidateStructureElementUtil.validateStructureElementInputCreate(structureElement, namingConvention);

        // validate data
        ValidateStructureElementUtil.validateStructureElementDataCreate(structureElement, namingConvention, holderRepositories, holderStructure);
    }
    public void validateStructuresCreate(List<StructureElementCommand> structureElements) {
        // initiate holder of system and device structure content, for performance reasons
        HolderSystemDeviceStructure holderStructure = new HolderSystemDeviceStructure(holderIRepositories);

        for (StructureElementCommand structureElement : structureElements) {
            validateStructuresCreate(structureElement, holderStructure);
        }
    }

    public void validateStructuresUpdate(StructureElementCommand structureElement) {
        validateStructuresUpdate(structureElement, new HolderSystemDeviceStructure(holderIRepositories));
    }
    public void validateStructuresUpdate(StructureElementCommand structureElement, HolderSystemDeviceStructure holderStructure) {
        // validate structure element
        //     input
        //         type
        //         input itself
        //     data
        //         data itself
        //         relative other data

        // validate input
        ValidateStructureElementUtil.validateStructureElementInputUpdate(structureElement, namingConvention);

        // validate data
        ValidateStructureElementUtil.validateStructureElementDataUpdate(structureElement, namingConvention, holderRepositories, holderStructure);
    }
    public void validateStructuresUpdate(List<StructureElementCommand> structureElements) {
        // initiate holder of system and device structure content, for performance reasons
        HolderSystemDeviceStructure holderStructure = new HolderSystemDeviceStructure(holderIRepositories);

        for (StructureElementCommand structureElement : structureElements) {
            validateStructuresUpdate(structureElement, holderStructure);
        }
    }

    public void validateStructuresDelete(StructureElementCommand structureElement) {
        validateStructuresDelete(structureElement, new HolderSystemDeviceStructure(holderIRepositories));
    }
    public void validateStructuresDelete(StructureElementCommand structureElement, HolderSystemDeviceStructure holderStructure) {
        // validate structure element
        //     input
        //         type
        //         input itself
        //     data
        //         data itself
        //         relative other data

        // validate input
        ValidateStructureElementUtil.validateStructureElementInputDelete(structureElement, namingConvention);

        // validate data
        ValidateStructureElementUtil.validateStructureElementDataDelete(structureElement, namingConvention, holderRepositories, holderStructure);
    }
    public void validateStructuresDelete(List<StructureElementCommand> structureElements) {
        // initiate holder of system and device structure content, for performance reasons
        HolderSystemDeviceStructure holderStructure = new HolderSystemDeviceStructure(holderIRepositories);

        for (StructureElementCommand structureElement : structureElements) {
            validateStructuresDelete(structureElement, holderStructure);
        }
    }

    public void validateStructuresApprove(StructureElementCommand structureElement) {
        validateStructuresApprove(structureElement, new HolderSystemDeviceStructure(holderIRepositories));
    }
    public void validateStructuresApprove(StructureElementCommand structureElement, HolderSystemDeviceStructure holderStructure) {
        // validate structure element
        //     input
        //         type
        //         input itself
        //     data
        //         data itself
        //         relative other data

        // validate input
        ValidateStructureElementUtil.validateStructureElementInputApprove(structureElement, namingConvention);

        // validate data
        ValidateStructureElementUtil.validateStructureElementDataApprove(structureElement, namingConvention, holderRepositories, holderStructure);
    }
    public void validateStructuresApprove(List<StructureElementCommand> structureElements) {
        // initiate holder of system and device structure content, for performance reasons
        HolderSystemDeviceStructure holderStructure = new HolderSystemDeviceStructure(holderIRepositories);

        for (StructureElementCommand structureElement : structureElements) {
            validateStructuresApprove(structureElement, holderStructure);
        }
    }

    public void validateStructuresCancel(StructureElementCommand structureElement) {
        validateStructuresCancel(structureElement, new HolderSystemDeviceStructure(holderIRepositories));
    }
    public void validateStructuresCancel(StructureElementCommand structureElement, HolderSystemDeviceStructure holderStructure) {
        // validate structure element
        //     input
        //         type
        //         input itself
        //     data
        //         data itself
        //         relative other data

        // validate input
        ValidateStructureElementUtil.validateStructureElementInputCancel(structureElement, namingConvention);

        // validate data
        ValidateStructureElementUtil.validateStructureElementDataCancel(structureElement, namingConvention, holderRepositories, holderStructure);
    }
    public void validateStructuresCancel(List<StructureElementCommand> structureElements) {
        // initiate holder of system and device structure content, for performance reasons
        HolderSystemDeviceStructure holderStructure = new HolderSystemDeviceStructure(holderIRepositories);

        for (StructureElementCommand structureElement : structureElements) {
            validateStructuresCancel(structureElement, holderStructure);
        }
    }

    public void validateStructuresReject(StructureElementCommand structureElement) {
        validateStructuresReject(structureElement, new HolderSystemDeviceStructure(holderIRepositories));
    }
    public void validateStructuresReject(StructureElementCommand structureElement, HolderSystemDeviceStructure holderStructure) {
        // validate structure element
        //     input
        //         type
        //         input itself
        //     data
        //         data itself
        //         relative other data

        // validate input
        ValidateStructureElementUtil.validateStructureElementInputReject(structureElement, namingConvention);

        // validate data
        ValidateStructureElementUtil.validateStructureElementDataReject(structureElement, namingConvention, holderRepositories, holderStructure);
    }
    public void validateStructuresReject(List<StructureElementCommand> structureElements) {
        // initiate holder of system and device structure content, for performance reasons
        HolderSystemDeviceStructure holderStructure = new HolderSystemDeviceStructure(holderIRepositories);

        for (StructureElementCommand structureElement : structureElements) {
            validateStructuresReject(structureElement, holderStructure);
        }
    }

    // ----------------------------------------------------------------------------------------------------

    @Transactional
    public List<StructureElement> updateStructures(List<StructureElementCommand> structureElements, String username) {
        // validation outside method
        // transaction
        //     for each structure element command
        //         update structure
        //         handle
        //             notification
        //             structure element for updated structure
        //     notify
        //     return structure elements for updated structures

        // initiate holder of system and device structure content, for performance reasons
        HolderSystemDeviceStructure holderStructure = new HolderSystemDeviceStructure(holderIRepositories);

        Date when = new Date();
        List<NotificationStructure> notifications = Lists.newArrayList();
        final List<StructureElement> updatedStructureElements = Lists.newArrayList();
        for (StructureElementCommand structureElement : structureElements) {
            Type type = structureElement.getType();
            StructureElementNotification structureElementNotification = null;

            if (Type.SYSTEMGROUP.equals(type)) {
                structureElementNotification = systemGroupService.updateStructure(structureElement, when, username, namingConvention, holderStructure);
            } else if (Type.SYSTEM.equals(type)) {
                structureElementNotification = systemService.updateStructure(structureElement, when, username, namingConvention, holderStructure);
            } else if (Type.SUBSYSTEM.equals(type)) {
                structureElementNotification = subsystemService.updateStructure(structureElement, when, username, namingConvention, holderStructure);
            } else if (Type.DISCIPLINE.equals(type)) {
                structureElementNotification = disciplineService.updateStructure(structureElement, when, username, namingConvention, holderStructure);
            } else if (Type.DEVICEGROUP.equals(type)) {
                structureElementNotification = deviceGroupService.updateStructure(structureElement, when, username, namingConvention, holderStructure);
            } else if (Type.DEVICETYPE.equals(type)) {
                structureElementNotification = deviceTypeService.updateStructure(structureElement, when, username, namingConvention, holderStructure);
            } else {
                continue;
            }

            Utilities.addToCollection(notifications, structureElementNotification.notificationStructure());
            Utilities.addToCollection(updatedStructureElements, structureElementNotification.structureElement());

            if (LOGGER.isLoggable(Level.FINER)) {
                LOGGER.log(Level.FINER, MessageFormat.format(TextUtil.DESCRIPTION_NAME_VALUE, TextUtil.UPDATE_STRUCTURE, "type", type));
                LOGGER.log(Level.FINER, MessageFormat.format(TextUtil.DESCRIPTION_NAME_VALUE, TextUtil.UPDATE_STRUCTURE, TextUtil.ELEMENT_IN, structureElement));
                LOGGER.log(Level.FINER, MessageFormat.format(TextUtil.DESCRIPTION_NAME_VALUE, TextUtil.UPDATE_STRUCTURE, TextUtil.ELEMENT_OUT, structureElementNotification.structureElement()));
                LOGGER.log(Level.FINER, MessageFormat.format(TextUtil.DESCRIPTION_NAME_VALUE, TextUtil.UPDATE_STRUCTURE, TextUtil.NOTIFICATION, structureElementNotification.notificationStructure()));
            }
        }

        // notify
        notificationService.notifyStructures(notifications, StructureCommand.UPDATE);

        LOGGER.log(Level.INFO,
                () -> MessageFormat.format(
                        TextUtil.DESCRIPTION_NUMBER_ELEMENTS_IN_OUT,
                        "Update structures",
                        structureElements.size(),
                        updatedStructureElements.size()));
        return updatedStructureElements;
    }

    // ----------------------------------------------------------------------------------------------------

    @Transactional
    public List<StructureElement> deleteStructures(List<StructureElementCommand> structureElements, String username) {
        // validation outside method
        // transaction
        //     for each structure element command
        //         delete structure
        //         handle
        //             notification
        //             structure element for deleted structure
        //     notify
        //     return structure elements for deleted structures

        // initiate holder of system and device structure content, for performance reasons
        HolderSystemDeviceStructure holderStructure = new HolderSystemDeviceStructure(holderIRepositories);

        Date when = new Date();
        List<NotificationStructure> notifications = Lists.newArrayList();
        final List<StructureElement> deletedStructureElements = Lists.newArrayList();
        for (StructureElementCommand structureElement : structureElements) {
            Type type = structureElement.getType();
            StructureElementNotification structureElementNotification = null;

            if (Type.SYSTEMGROUP.equals(type)) {
                structureElementNotification = systemGroupService.deleteStructure(structureElement, when, username, holderStructure);
            } else if (Type.SYSTEM.equals(type)) {
                structureElementNotification = systemService.deleteStructure(structureElement, when, username, holderStructure);
            } else if (Type.SUBSYSTEM.equals(type)) {
                structureElementNotification = subsystemService.deleteStructure(structureElement, when, username, holderStructure);
            } else if (Type.DISCIPLINE.equals(type)) {
                structureElementNotification = disciplineService.deleteStructure(structureElement, when, username, holderStructure);
            } else if (Type.DEVICEGROUP.equals(type)) {
                structureElementNotification = deviceGroupService.deleteStructure(structureElement, when, username, holderStructure);
            } else if (Type.DEVICETYPE.equals(type)) {
                structureElementNotification = deviceTypeService.deleteStructure(structureElement, when, username, holderStructure);
            } else {
                continue;
            }

            Utilities.addToCollection(notifications, structureElementNotification.notificationStructure());
            Utilities.addToCollection(deletedStructureElements, structureElementNotification.structureElement());

            if (LOGGER.isLoggable(Level.FINER)) {
                LOGGER.log(Level.FINER, MessageFormat.format(TextUtil.DESCRIPTION_NAME_VALUE, TextUtil.DELETE_STRUCTURE, "type", type));
                LOGGER.log(Level.FINER, MessageFormat.format(TextUtil.DESCRIPTION_NAME_VALUE, TextUtil.DELETE_STRUCTURE, TextUtil.ELEMENT_IN, structureElement));
                LOGGER.log(Level.FINER, MessageFormat.format(TextUtil.DESCRIPTION_NAME_VALUE, TextUtil.DELETE_STRUCTURE, TextUtil.ELEMENT_OUT, structureElementNotification.structureElement()));
                LOGGER.log(Level.FINER, MessageFormat.format(TextUtil.DESCRIPTION_NAME_VALUE, TextUtil.DELETE_STRUCTURE, TextUtil.NOTIFICATION, structureElementNotification.notificationStructure()));
            }
        }

        // notify
        notificationService.notifyStructures(notifications, StructureCommand.DELETE);

        LOGGER.log(Level.INFO,
                () -> MessageFormat.format(
                        TextUtil.DESCRIPTION_NUMBER_ELEMENTS_IN_OUT,
                        "Delete structures",
                        structureElements.size(),
                        deletedStructureElements.size()));
        return deletedStructureElements;
    }

    // ----------------------------------------------------------------------------------------------------

    @Transactional
    public List<StructureElement> approveStructures(List<StructureElementCommand> structureElements, String username) {
        // validation outside method
        // transaction
        //     for each structure element command
        //         approve structure
        //         handle
        //             notification
        //             structure element for approved structure
        //     notify
        //     return structure elements for approved structures

        // initiate holder of system and device structure content, for performance reasons
        HolderSystemDeviceStructure holderStructure = new HolderSystemDeviceStructure(holderIRepositories);

        // approve
        //     set not latest for current latest
        //     set approved, latest for pending
        //     delete
        //         if deleted, names that belong to structure may be deleted, otherwise they are alive but become legacy names - legacy way for now
        //         delete sub structures

        Date when = new Date();
        List<NotificationStructure> notifications = Lists.newArrayList();
        final List<StructureElement> approvedStructureElements = Lists.newArrayList();
        for (StructureElementCommand structureElement : structureElements) {
            Type type = structureElement.getType();
            StructureElementNotification structureElementNotification = null;

            if (Type.SYSTEMGROUP.equals(type)) {
                structureElementNotification = systemGroupService.approveStructure(structureElement, when, username, holderStructure);
            } else if (Type.SYSTEM.equals(type)) {
                structureElementNotification = systemService.approveStructure(structureElement, when, username, holderStructure);
            } else if (Type.SUBSYSTEM.equals(type)) {
                structureElementNotification = subsystemService.approveStructure(structureElement, when, username, holderStructure);
            } else if (Type.DISCIPLINE.equals(type)) {
                structureElementNotification = disciplineService.approveStructure(structureElement, when, username, holderStructure);
            } else if (Type.DEVICEGROUP.equals(type)) {
                structureElementNotification = deviceGroupService.approveStructure(structureElement, when, username, holderStructure);
            } else if (Type.DEVICETYPE.equals(structureElement.getType())) {
                structureElementNotification = deviceTypeService.approveStructure(structureElement, when, username, holderStructure);
            } else {
                continue;
            }

            Utilities.addToCollection(notifications, structureElementNotification.notificationStructure());
            Utilities.addToCollection(approvedStructureElements, structureElementNotification.structureElement());

            if (LOGGER.isLoggable(Level.FINER)) {
                LOGGER.log(Level.FINER, MessageFormat.format(TextUtil.DESCRIPTION_NAME_VALUE, TextUtil.APPROVE_STRUCTURE, "type", type));
                LOGGER.log(Level.FINER, MessageFormat.format(TextUtil.DESCRIPTION_NAME_VALUE, TextUtil.APPROVE_STRUCTURE, TextUtil.ELEMENT_IN, structureElement));
                LOGGER.log(Level.FINER, MessageFormat.format(TextUtil.DESCRIPTION_NAME_VALUE, TextUtil.APPROVE_STRUCTURE, TextUtil.ELEMENT_OUT, structureElementNotification.structureElement()));
                LOGGER.log(Level.FINER, MessageFormat.format(TextUtil.DESCRIPTION_NAME_VALUE, TextUtil.APPROVE_STRUCTURE, TextUtil.NOTIFICATION, structureElementNotification.notificationStructure()));
            }
        }

        // notify
        notificationService.notifyStructures(notifications, StructureCommand.APPROVE);

        LOGGER.log(Level.INFO,
                () -> MessageFormat.format(
                        TextUtil.DESCRIPTION_NUMBER_ELEMENTS_IN_OUT,
                        "Approve structures",
                        structureElements.size(),
                        approvedStructureElements.size()));
        return approvedStructureElements;
    }

    @Transactional
    public List<StructureElement> cancelStructures(List<StructureElementCommand> structureElements, String username) {
        // validation outside method
        // transaction
        //     for each structure element command
        //         cancel structure
        //         handle
        //             notification
        //             structure element for cancelled structure
        //     notify
        //     return structure elements for cancelled structures

        // initiate holder of system and device structure content, for performance reasons
        HolderSystemDeviceStructure holderStructure = new HolderSystemDeviceStructure(holderIRepositories);

        Date when = new Date();
        List<NotificationStructure> notifications = Lists.newArrayList();
        final List<StructureElement> cancelledStructureElements = Lists.newArrayList();
        for (StructureElementCommand structureElement : structureElements) {
            Type type = structureElement.getType();
            StructureElementNotification structureElementNotification = null;

            if (Type.SYSTEMGROUP.equals(type)) {
                structureElementNotification = systemGroupService.cancelStructure(structureElement, when, username, holderStructure);
            } else if (Type.SYSTEM.equals(type)) {
                structureElementNotification = systemService.cancelStructure(structureElement, when, username, holderStructure);
            } else if (Type.SUBSYSTEM.equals(type)) {
                structureElementNotification = subsystemService.cancelStructure(structureElement, when, username, holderStructure);
            } else if (Type.DISCIPLINE.equals(type)) {
                structureElementNotification = disciplineService.cancelStructure(structureElement, when, username, holderStructure);
            } else if (Type.DEVICEGROUP.equals(type)) {
                structureElementNotification = deviceGroupService.cancelStructure(structureElement, when, username, holderStructure);
            } else if (Type.DEVICETYPE.equals(structureElement.getType())) {
                structureElementNotification = deviceTypeService.cancelStructure(structureElement, when, username, holderStructure);
            } else {
                continue;
            }

            Utilities.addToCollection(notifications, structureElementNotification.notificationStructure());
            Utilities.addToCollection(cancelledStructureElements, structureElementNotification.structureElement());

            if (LOGGER.isLoggable(Level.FINER)) {
                LOGGER.log(Level.FINER, MessageFormat.format(TextUtil.DESCRIPTION_NAME_VALUE, TextUtil.CANCEL_STRUCTURE, "type", type));
                LOGGER.log(Level.FINER, MessageFormat.format(TextUtil.DESCRIPTION_NAME_VALUE, TextUtil.CANCEL_STRUCTURE, TextUtil.ELEMENT_IN, structureElement));
                LOGGER.log(Level.FINER, MessageFormat.format(TextUtil.DESCRIPTION_NAME_VALUE, TextUtil.CANCEL_STRUCTURE, TextUtil.ELEMENT_OUT, structureElementNotification.structureElement()));
                LOGGER.log(Level.FINER, MessageFormat.format(TextUtil.DESCRIPTION_NAME_VALUE, TextUtil.CANCEL_STRUCTURE, TextUtil.NOTIFICATION, structureElementNotification.notificationStructure()));
            }
        }

        // notify
        notificationService.notifyStructures(notifications, StructureCommand.CANCEL);

        LOGGER.log(Level.INFO,
                () -> MessageFormat.format(
                        TextUtil.DESCRIPTION_NUMBER_ELEMENTS_IN_OUT,
                        "Cancel structures",
                        structureElements.size(),
                        cancelledStructureElements.size()));
        return cancelledStructureElements;
    }

    @Transactional
    public List<StructureElement> rejectStructures(List<StructureElementCommand> structureElements, String username) {
        // validation outside method
        // transaction
        //     for each structure element command
        //         reject structure
        //         handle
        //             notification
        //             structure element for rejected structure
        //     notify
        //     return structure elements for rejected structures

        // initiate holder of system and device structure content, for performance reasons
        HolderSystemDeviceStructure holderStructure = new HolderSystemDeviceStructure(holderIRepositories);

        Date when = new Date();
        List<NotificationStructure> notifications = Lists.newArrayList();
        final List<StructureElement> rejectedStructureElements = Lists.newArrayList();
        for (StructureElementCommand structureElement : structureElements) {
            Type type = structureElement.getType();
            StructureElementNotification structureElementNotification = null;

            if (Type.SYSTEMGROUP.equals(type)) {
                structureElementNotification = systemGroupService.rejectStructure(structureElement, when, username, holderStructure);
            } else if (Type.SYSTEM.equals(type)) {
                structureElementNotification = systemService.rejectStructure(structureElement, when, username, holderStructure);
            } else if (Type.SUBSYSTEM.equals(type)) {
                structureElementNotification = subsystemService.rejectStructure(structureElement, when, username, holderStructure);
            } else if (Type.DISCIPLINE.equals(type)) {
                structureElementNotification = disciplineService.rejectStructure(structureElement, when, username, holderStructure);
            } else if (Type.DEVICEGROUP.equals(type)) {
                structureElementNotification = deviceGroupService.rejectStructure(structureElement, when, username, holderStructure);
            } else if (Type.DEVICETYPE.equals(structureElement.getType())) {
                structureElementNotification = deviceTypeService.rejectStructure(structureElement, when, username, holderStructure);
            } else {
                continue;
            }

            Utilities.addToCollection(notifications, structureElementNotification.notificationStructure());
            Utilities.addToCollection(rejectedStructureElements, structureElementNotification.structureElement());

            if (LOGGER.isLoggable(Level.FINER)) {
                LOGGER.log(Level.FINER, MessageFormat.format(TextUtil.DESCRIPTION_NAME_VALUE, TextUtil.REJECT_STRUCTURE, "type", type));
                LOGGER.log(Level.FINER, MessageFormat.format(TextUtil.DESCRIPTION_NAME_VALUE, TextUtil.REJECT_STRUCTURE, TextUtil.ELEMENT_IN, structureElement));
                LOGGER.log(Level.FINER, MessageFormat.format(TextUtil.DESCRIPTION_NAME_VALUE, TextUtil.REJECT_STRUCTURE, TextUtil.ELEMENT_OUT, structureElementNotification.structureElement()));
                LOGGER.log(Level.FINER, MessageFormat.format(TextUtil.DESCRIPTION_NAME_VALUE, TextUtil.REJECT_STRUCTURE, TextUtil.NOTIFICATION, structureElementNotification.notificationStructure()));
            }
        }

        // notify
        notificationService.notifyStructures(notifications, StructureCommand.REJECT);

        LOGGER.log(Level.INFO,
                () -> MessageFormat.format(
                        TextUtil.DESCRIPTION_NUMBER_ELEMENTS_IN_OUT,
                        "Reject structures",
                        structureElements.size(),
                        rejectedStructureElements.size()));
        return rejectedStructureElements;
    }

}
