/*
 * Copyright (C) 2021 European Spallation Source ERIC.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package org.openepics.names.util;

import java.util.Date;
import java.util.List;
import java.util.UUID;

import org.openepics.names.repository.model.DeviceGroup;
import org.openepics.names.repository.model.DeviceType;
import org.openepics.names.repository.model.Discipline;
import org.openepics.names.repository.model.NameStructure;
import org.openepics.names.repository.model.Structure;
import org.openepics.names.repository.model.Subsystem;
import org.openepics.names.repository.model.System;
import org.openepics.names.repository.model.SystemGroup;
import org.openepics.names.rest.beans.Status;
import org.openepics.names.rest.beans.Type;
import org.openepics.names.rest.beans.element.StructureElement;
import org.openepics.names.rest.beans.element.StructureElementCommand;
import org.openepics.names.rest.beans.element.StructureElementCommandConfirm;
import org.openepics.names.rest.beans.element.StructureElementCommandCreate;
import org.openepics.names.rest.beans.element.StructureElementCommandUpdate;

import com.google.common.collect.Lists;

/**
 * Utility class to assist in populating structure elements based on repository content.
 *
 * @author Lars Johansson
 */
public class StructureElementUtil {

    // used in handling structure attributes - requested, processed
    private static final long THOUSAND_MILLISECONDS = 1000;

    /**
     * This class is not to be instantiated.
     */
    private StructureElementUtil() {
        throw new IllegalStateException("Utility class");
    }

    /**
     * Populate and return list of structure elements for system groups.
     *
     * @param systemGroups system groups
     * @param holderStructure holder of system and device structure content
     * @param structureChoice whether to consider content from structure perspective or history perspective.
     *                        Structure perspective gives one StructureElement object (processed).
     *                        History perspective gives two StructureElement objects (requested, processed).
     *                        If choice not given then default as structure perspective (processed).
     * @return list of structure elements
     */
    public static List<StructureElement> getStructureElementsForSystemGroups(List<SystemGroup> systemGroups, HolderSystemDeviceStructure holderStructure, StructureChoice structureChoice) {
        List<StructureElement> structureElements = Lists.newArrayList();
        for (SystemGroup systemGroup : systemGroups) {
            // one or two return elements
            //     processed != null and processed != requested (>  1s difference) --> two entries (requested, processed)
            //     processed != null and processed == requested (<= 1s difference) --> one entry   (processed initial)
            //     processed == null                                               --> one entry   (requested)

            if (StructureChoice.HISTORY.equals(structureChoice)) {
                if (systemGroup.getProcessed() != null && ((systemGroup.getProcessed().getTime() - systemGroup.getRequested().getTime()) > THOUSAND_MILLISECONDS)) {
                    structureElements.add(getStructureElementRequested(systemGroup));
                    structureElements.add(getStructureElementProcessed(systemGroup, holderStructure, structureChoice));
                } else if (systemGroup.getProcessed() != null && ((systemGroup.getProcessed().getTime() - systemGroup.getRequested().getTime()) <= THOUSAND_MILLISECONDS)) {
                    structureElements.add(getStructureElementProcessed(systemGroup, holderStructure, structureChoice));
                } else {
                    structureElements.add(getStructureElementRequested(systemGroup));
                }
            } else {
                if (Status.PENDING.equals(systemGroup.getStatus())) {
                    structureElements.add(getStructureElementRequested(systemGroup));
                } else {
                    structureElements.add(getStructureElementProcessed(systemGroup, holderStructure, structureChoice));
                }
            }
        }
        return structureElements;
    }
    /**
     * Populate and return list of structure elements for systems.
     *
     * @param systems systems
     * @param holderStructure holder of system and device structure content
     * @param structureChoice whether to consider content from structure perspective or history perspective.
     *                        Structure perspective gives one StructureElement object (processed).
     *                        History perspective gives two StructureElement objects (requested, processed).
     *                        If choice not given then default as structure perspective (processed).
     * @return list of structure elements
     */
    public static List<StructureElement> getStructureElementsForSystems(List<System> systems, HolderSystemDeviceStructure holderStructure, StructureChoice structureChoice) {
        List<StructureElement> structureElements = Lists.newArrayList();
        for (System system : systems) {
            // one or two return elements
            //     processed != null and processed != requested (>  1s difference) --> two entries (requested, processed)
            //     processed != null and processed == requested (<= 1s difference) --> one entry   (processed initial)
            //     processed == null                                               --> one entry   (requested)

            if (StructureChoice.HISTORY.equals(structureChoice)) {
                if (system.getProcessed() != null && ((system.getProcessed().getTime() - system.getRequested().getTime()) > THOUSAND_MILLISECONDS)) {
                    structureElements.add(getStructureElementRequested(system));
                    structureElements.add(getStructureElementProcessed(system, holderStructure, structureChoice));
                } else if (system.getProcessed() != null && ((system.getProcessed().getTime() - system.getRequested().getTime()) <= THOUSAND_MILLISECONDS)) {
                    structureElements.add(getStructureElementProcessed(system, holderStructure, structureChoice));
                } else {
                    structureElements.add(getStructureElementRequested(system));
                }
            } else {
                if (Status.PENDING.equals(system.getStatus())) {
                    structureElements.add(getStructureElementRequested(system));
                } else {
                    structureElements.add(getStructureElementProcessed(system, holderStructure, structureChoice));
                }
            }
        }
        return structureElements;
    }
    /**
     * Populate and return list of structure elements for subsystems.
     *
     * @param subsystems subsystems
     * @param holderStructure holder of system and device structure content
     * @param structureChoice whether to consider content from structure perspective or history perspective.
     *                        Structure perspective gives one StructureElement object (processed).
     *                        History perspective gives two StructureElement objects (requested, processed).
     *                        If choice not given then default as structure perspective (processed).
     * @return list of structure elements
     */
    public static List<StructureElement> getStructureElementsForSubsystems(List<Subsystem> subsystems, HolderSystemDeviceStructure holderStructure, StructureChoice structureChoice) {
        List<StructureElement> structureElements = Lists.newArrayList();
        for (Subsystem subsystem : subsystems) {
            // one or two return elements
            //     processed != null and processed != requested (>  1s difference) --> two entries (requested, processed)
            //     processed != null and processed == requested (<= 1s difference) --> one entry   (processed initial)
            //     processed == null                                               --> one entry   (requested)

            if (StructureChoice.HISTORY.equals(structureChoice)) {
                if (subsystem.getProcessed() != null && ((subsystem.getProcessed().getTime() - subsystem.getRequested().getTime()) > THOUSAND_MILLISECONDS)) {
                    structureElements.add(getStructureElementRequested(subsystem));
                    structureElements.add(getStructureElementProcessed(subsystem, holderStructure, structureChoice));
                } else if (subsystem.getProcessed() != null && ((subsystem.getProcessed().getTime() - subsystem.getRequested().getTime()) <= THOUSAND_MILLISECONDS)) {
                    structureElements.add(getStructureElementProcessed(subsystem, holderStructure, structureChoice));
                } else {
                    structureElements.add(getStructureElementRequested(subsystem));
                }
            } else {
                if (Status.PENDING.equals(subsystem.getStatus())) {
                    structureElements.add(getStructureElementRequested(subsystem));
                } else {
                    structureElements.add(getStructureElementProcessed(subsystem, holderStructure, structureChoice));
                }
            }
        }
        return structureElements;
    }

    /**
     * Populate and return list of structure elements for disciplines.
     *
     * @param disciplines disciplines
     * @param holderStructure holder of system and device structure content
     * @param structureChoice whether to consider content from structure perspective or history perspective.
     *                        Structure perspective gives one StructureElement object (processed).
     *                        History perspective gives two StructureElement objects (requested, processed).
     *                        If choice not given then default as structure perspective (processed).
     * @return list of structure elements
     */
    public static List<StructureElement> getStructureElementsForDisciplines(List<Discipline> disciplines, HolderSystemDeviceStructure holderStructure, StructureChoice structureChoice) {
        List<StructureElement> structureElements = Lists.newArrayList();
        for (Discipline discipline : disciplines) {
            // one or two return elements
            //     processed != null and processed != requested (>  1s difference) --> two entries (requested, processed)
            //     processed != null and processed == requested (<= 1s difference) --> one entry   (processed initial)
            //     processed == null                                               --> one entry   (requested)

            if (StructureChoice.HISTORY.equals(structureChoice)) {
                if (discipline.getProcessed() != null && ((discipline.getProcessed().getTime() - discipline.getRequested().getTime()) > THOUSAND_MILLISECONDS)) {
                    structureElements.add(getStructureElementRequested(discipline));
                    structureElements.add(getStructureElementProcessed(discipline, holderStructure, structureChoice));
                } else if (discipline.getProcessed() != null && ((discipline.getProcessed().getTime() - discipline.getRequested().getTime()) <= THOUSAND_MILLISECONDS)) {
                    structureElements.add(getStructureElementProcessed(discipline, holderStructure, structureChoice));
                } else {
                    structureElements.add(getStructureElementRequested(discipline));
                }
            } else {
                if (Status.PENDING.equals(discipline.getStatus())) {
                    structureElements.add(getStructureElementRequested(discipline));
                } else {
                    structureElements.add(getStructureElementProcessed(discipline, holderStructure, structureChoice));
                }
            }
        }
        return structureElements;
    }
    /**
     * Populate and return list of structure elements for device groups.
     *
     * @param deviceGroups device groups
     * @param holderStructure holder of system and device structure content
     * @param structureChoice whether to consider content from structure perspective or history perspective.
     *                        Structure perspective gives one StructureElement object (processed).
     *                        History perspective gives two StructureElement objects (requested, processed).
     *                        If choice not given then default as structure perspective (processed).
     * @return list of structure elements
     */
    public static List<StructureElement> getStructureElementsForDeviceGroups(List<DeviceGroup> deviceGroups, HolderSystemDeviceStructure holderStructure, StructureChoice structureChoice) {
        List<StructureElement> structureElements = Lists.newArrayList();
        for (DeviceGroup deviceGroup : deviceGroups) {
            // one or two return elements
            //     processed != null and processed != requested (>  1s difference) --> two entries (requested, processed)
            //     processed != null and processed == requested (<= 1s difference) --> one entry   (processed initial)
            //     processed == null                                               --> one entry   (requested)

            if (StructureChoice.HISTORY.equals(structureChoice)) {
                if (deviceGroup.getProcessed() != null && ((deviceGroup.getProcessed().getTime() - deviceGroup.getRequested().getTime()) > THOUSAND_MILLISECONDS)) {
                    structureElements.add(getStructureElementRequested(deviceGroup));
                    structureElements.add(getStructureElementProcessed(deviceGroup, holderStructure, structureChoice));
                } else if (deviceGroup.getProcessed() != null && ((deviceGroup.getProcessed().getTime() - deviceGroup.getRequested().getTime()) <= THOUSAND_MILLISECONDS)) {
                    structureElements.add(getStructureElementProcessed(deviceGroup, holderStructure, structureChoice));
                } else {
                    structureElements.add(getStructureElementRequested(deviceGroup));
                }
            } else {
                if (Status.PENDING.equals(deviceGroup.getStatus())) {
                    structureElements.add(getStructureElementRequested(deviceGroup));
                } else {
                    structureElements.add(getStructureElementProcessed(deviceGroup, holderStructure, structureChoice));
                }
            }
        }
        return structureElements;
    }
    /**
     * Populate and return list of structure elements for device types.
     *
     * @param deviceTypes device types
     * @param holderStructure holder of system and device structure content
     * @param structureChoice whether to consider content from structure perspective or history perspective.
     *                        Structure perspective gives one StructureElement object (processed).
     *                        History perspective gives two StructureElement objects (requested, processed).
     *                        If choice not given then default as structure perspective (processed).
     * @return list of structure elements
     */
    public static List<StructureElement> getStructureElementsForDeviceTypes(List<DeviceType> deviceTypes, HolderSystemDeviceStructure holderStructure, StructureChoice structureChoice) {
        List<StructureElement> structureElements = Lists.newArrayList();
        for (DeviceType deviceType : deviceTypes) {
            // one or two return elements
            //     processed != null and processed != requested (>  1s difference) --> two entries (requested, processed)
            //     processed != null and processed == requested (<= 1s difference) --> one entry   (processed initial)
            //     processed == null                                               --> one entry   (requested)

            if (StructureChoice.HISTORY.equals(structureChoice)) {
                if (deviceType.getProcessed() != null && ((deviceType.getProcessed().getTime() - deviceType.getRequested().getTime()) > THOUSAND_MILLISECONDS)) {
                    structureElements.add(getStructureElementRequested(deviceType));
                    structureElements.add(getStructureElementProcessed(deviceType, holderStructure, structureChoice));
                } else if (deviceType.getProcessed() != null && ((deviceType.getProcessed().getTime() - deviceType.getRequested().getTime()) <= THOUSAND_MILLISECONDS)) {
                    structureElements.add(getStructureElementProcessed(deviceType, holderStructure, structureChoice));
                } else {
                    structureElements.add(getStructureElementRequested(deviceType));
                }
            } else {
                if (Status.PENDING.equals(deviceType.getStatus())) {
                    structureElements.add(getStructureElementRequested(deviceType));
                } else {
                    structureElements.add(getStructureElementProcessed(deviceType, holderStructure, structureChoice));
                }
            }
        }
        return structureElements;
    }

    /**
     * Populate and return structure element for system group with focus on requested.
     *
     * @param systemGroup system group
     * @return structure element
     */
    public static StructureElement getStructureElementRequested(SystemGroup systemGroup) {
        if (systemGroup == null) {
            return null;
        }

        // mnemonic path
        //     ambiguous if considered in past (requested, processed)
        //     not for history of structure
        //     for latest & approved status
        String mnemonicPath = null;

        return getStructureElement(
                systemGroup.getUuid(),
                Type.SYSTEMGROUP,
                null,
                systemGroup.getName(), systemGroup.getMnemonic(), mnemonicPath, 1,
                systemGroup.getDescription(), Status.PENDING, Boolean.FALSE, systemGroup.isDeleted(),
                systemGroup.getRequested(), systemGroup.getRequestedBy(), systemGroup.getRequestedComment());
    }
    /**
     * Populate and return structure element for system group with focus on processed.
     *
     * @param systemGroup system group
     * @param holderStructure holder of system and device structure content
     * @param structureChoice whether to consider content from structure perspective or history perspective
     * @return structure element
     */
    public static StructureElement getStructureElementProcessed(SystemGroup systemGroup, HolderSystemDeviceStructure holderStructure, StructureChoice structureChoice) {
        if (systemGroup == null) {
            return null;
        }

        // mnemonic path
        //     ambiguous if considered in past (requested, processed)
        //     not history of structure
        //     latest & approved status
        String mnemonicPath = StructureChoice.HISTORY.equals(structureChoice) || !(systemGroup.isLatest() && Status.APPROVED.equals(systemGroup.getStatus()))
                ? null
                : StructureUtil.getMnemonicPath(systemGroup, holderStructure);

        return getStructureElement(
                systemGroup.getUuid(),
                Type.SYSTEMGROUP,
                null,
                systemGroup.getName(), systemGroup.getMnemonic(), mnemonicPath, 1,
                systemGroup.getDescription(), systemGroup.getStatus(), systemGroup.isLatest(), systemGroup.isDeleted(),
                systemGroup.getProcessed(), systemGroup.getProcessedBy(), systemGroup.getProcessedComment());
    }
    /**
     * Populate and return structure element for system group with focus on requested.
     *
     * @param system system
     * @return structure element
     */
    public static StructureElement getStructureElementRequested(System system) {
        if (system == null) {
            return null;
        }

        // mnemonic path
        //     ambiguous if considered in past (requested, processed)
        //     not history of structure
        //     latest & approved status
        String mnemonicPath = null;

        return getStructureElement(
                system.getUuid(),
                Type.SYSTEM,
                system.getParentUuid(),
                system.getName(), system.getMnemonic(), mnemonicPath, 2,
                system.getDescription(), Status.PENDING, Boolean.FALSE, system.isDeleted(),
                system.getRequested(), system.getRequestedBy(), system.getRequestedComment());
    }
    /**
     * Populate and return structure element for system with focus on processed.
     *
     * @param system system
     * @param holderStructure holder of system and device structure content
     * @param structureChoice whether to consider content from structure perspective or history perspective
     * @return structure element
     */
    public static StructureElement getStructureElementProcessed(System system, HolderSystemDeviceStructure holderStructure, StructureChoice structureChoice) {
        if (system == null) {
            return null;
        }

        // mnemonic path
        //     ambiguous if considered in past (requested, processed)
        //     not history of structure
        //     latest & approved status
        String mnemonicPath = StructureChoice.HISTORY.equals(structureChoice) || !(system.isLatest() && Status.APPROVED.equals(system.getStatus()))
                ? null
                : StructureUtil.getMnemonicPath(system, holderStructure);

        return getStructureElement(
                system.getUuid(),
                Type.SYSTEM,
                system.getParentUuid(),
                system.getName(), system.getMnemonic(), mnemonicPath, 2,
                system.getDescription(), system.getStatus(), system.isLatest(), system.isDeleted(),
                system.getProcessed(), system.getProcessedBy(), system.getProcessedComment());
    }
    /**
     * Populate and return structure element for subsystem with focus on requested.
     *
     * @param subsystem subsystem
     * @return structure element
     */
    public static StructureElement getStructureElementRequested(Subsystem subsystem) {
        if (subsystem == null) {
            return null;
        }

        // mnemonic path
        //     ambiguous if considered in past (requested, processed)
        //     not history of structure
        //     latest & approved status
        String mnemonicPath = null;

        return getStructureElement(
                subsystem.getUuid(),
                Type.SUBSYSTEM,
                subsystem.getParentUuid(),
                subsystem.getName(), subsystem.getMnemonic(), mnemonicPath, 3,
                subsystem.getDescription(), Status.PENDING, Boolean.FALSE, subsystem.isDeleted(),
                subsystem.getRequested(), subsystem.getRequestedBy(), subsystem.getRequestedComment());
    }
    /**
     * Populate and return structure element for subsystem with focus on processed.
     *
     * @param subsystem subsystem
     * @param holderStructure holder of system and device structure content
     * @param structureChoice whether to consider content from structure perspective or history perspective
     * @return structure element
     */
    public static StructureElement getStructureElementProcessed(Subsystem subsystem, HolderSystemDeviceStructure holderStructure, StructureChoice structureChoice) {
        if (subsystem == null) {
            return null;
        }

        // mnemonic path
        //     ambiguous if considered in past (requested, processed)
        //     not history of structure
        //     latest & approved status
        String mnemonicPath = StructureChoice.HISTORY.equals(structureChoice) || !(subsystem.isLatest() && Status.APPROVED.equals(subsystem.getStatus()))
                ? null
                : StructureUtil.getMnemonicPath(subsystem, holderStructure);

        return getStructureElement(
                subsystem.getUuid(),
                Type.SUBSYSTEM,
                subsystem.getParentUuid(),
                subsystem.getName(), subsystem.getMnemonic(), mnemonicPath, 3,
                subsystem.getDescription(), subsystem.getStatus(), subsystem.isLatest(), subsystem.isDeleted(),
                subsystem.getProcessed(), subsystem.getProcessedBy(), subsystem.getProcessedComment());
    }

    /**
     * Populate and return structure element for discipline with focus on requested.
     *
     * @param discipline discipline
     * @return structure element
     */
    public static StructureElement getStructureElementRequested(Discipline discipline) {
        if (discipline == null) {
            return null;
        }

        // mnemonic path
        //     ambiguous if considered in past (requested, processed)
        //     not history of structure
        //     latest & approved status
        String mnemonicPath = null;

        return getStructureElement(
                discipline.getUuid(),
                Type.DISCIPLINE,
                null,
                discipline.getName(), discipline.getMnemonic(), mnemonicPath, 1,
                discipline.getDescription(), Status.PENDING, Boolean.FALSE, discipline.isDeleted(),
                discipline.getRequested(), discipline.getRequestedBy(), discipline.getRequestedComment());
    }
    /**
     * Populate and return structure element for discipline with focus on processed.
     *
     * @param discipline discipline
     * @param holderStructure holder of system and device structure content
     * @param structureChoice whether to consider content from structure perspective or history perspective
     * @return structure element
     */
    public static StructureElement getStructureElementProcessed(Discipline discipline, HolderSystemDeviceStructure holderStructure, StructureChoice structureChoice) {
        if (discipline == null) {
            return null;
        }

        // mnemonic path
        //     ambiguous if considered in past (requested, processed)
        //     not history of structure
        //     latest & approved status
        String mnemonicPath = StructureChoice.HISTORY.equals(structureChoice) || !(discipline.isLatest() && Status.APPROVED.equals(discipline.getStatus()))
                ? null
                : StructureUtil.getMnemonicPath(discipline, holderStructure);

        return getStructureElement(
                discipline.getUuid(),
                Type.DISCIPLINE,
                null,
                discipline.getName(), discipline.getMnemonic(), mnemonicPath, 1,
                discipline.getDescription(), discipline.getStatus(), discipline.isLatest(), discipline.isDeleted(),
                discipline.getProcessed(), discipline.getProcessedBy(), discipline.getProcessedComment());
    }
    /**
     * Populate and return structure element for device group with focus on requested.
     *
     * @param deviceGroup device group
     * @return structure element
     */
    public static StructureElement getStructureElementRequested(DeviceGroup deviceGroup) {
        if (deviceGroup == null) {
            return null;
        }

        // mnemonic path
        //     ambiguous if considered in past (requested, processed)
        //     not history of structure
        //     latest & approved status
        String mnemonicPath = null;

        return getStructureElement(
                deviceGroup.getUuid(),
                Type.DEVICEGROUP,
                deviceGroup.getParentUuid(),
                deviceGroup.getName(), deviceGroup.getMnemonic(), mnemonicPath, 2,
                deviceGroup.getDescription(), Status.PENDING, Boolean.FALSE, deviceGroup.isDeleted(),
                deviceGroup.getRequested(), deviceGroup.getRequestedBy(), deviceGroup.getRequestedComment());
    }
    /**
     * Populate and return structure element for device group with focus on processed.
     *
     * @param deviceGroup device group
     * @param holderStructure holder of system and device structure content
     * @param structureChoice whether to consider content from structure perspective or history perspective
     * @return structure element
     */
    public static StructureElement getStructureElementProcessed(DeviceGroup deviceGroup, HolderSystemDeviceStructure holderStructure, StructureChoice structureChoice) {
        if (deviceGroup == null) {
            return null;
        }

        // mnemonic path
        //     ambiguous if considered in past (requested, processed)
        //     not history of structure
        //     latest & approved status
        String mnemonicPath = StructureChoice.HISTORY.equals(structureChoice) || !(deviceGroup.isLatest() && Status.APPROVED.equals(deviceGroup.getStatus()))
                ? null
                : StructureUtil.getMnemonicPath(deviceGroup, holderStructure);

        return getStructureElement(
                deviceGroup.getUuid(),
                Type.DEVICEGROUP,
                deviceGroup.getParentUuid(),
                deviceGroup.getName(), deviceGroup.getMnemonic(), mnemonicPath, 2,
                deviceGroup.getDescription(), deviceGroup.getStatus(), deviceGroup.isLatest(), deviceGroup.isDeleted(),
                deviceGroup.getProcessed(), deviceGroup.getProcessedBy(), deviceGroup.getProcessedComment());
    }
    /**
     * Populate and return structure element for device type with focus on requested.
     *
     * @param deviceType device type
     * @return  structure element
     */
    public static StructureElement getStructureElementRequested(DeviceType deviceType) {
        if (deviceType == null) {
            return null;
        }

        // mnemonic path
        //     ambiguous if considered in past (requested, processed)
        //     not history of structure
        //     latest & approved status
        String mnemonicPath = null;

        return getStructureElement(
                deviceType.getUuid(),
                Type.DEVICETYPE,
                deviceType.getParentUuid(),
                deviceType.getName(), deviceType.getMnemonic(), mnemonicPath, 3,
                deviceType.getDescription(), Status.PENDING, Boolean.FALSE, deviceType.isDeleted(),
                deviceType.getRequested(), deviceType.getRequestedBy(), deviceType.getRequestedComment());
    }
    /**
     * Populate and return structure element for device type with focus on processed.
     *
     * @param deviceType device type
     * @param holderStructure holder of system and device structure content
     * @param structureChoice whether to consider content from structure perspective or history perspective
     * @return structure element
     */
    public static StructureElement getStructureElementProcessed(DeviceType deviceType, HolderSystemDeviceStructure holderStructure, StructureChoice structureChoice) {
        if (deviceType == null) {
            return null;
        }

        // mnemonic path
        //     ambiguous if considered in past (requested, processed)
        //     not history of structure
        //     latest & approved status
        String mnemonicPath = StructureChoice.HISTORY.equals(structureChoice) || !(deviceType.isLatest() && Status.APPROVED.equals(deviceType.getStatus()))
                ? null
                : StructureUtil.getMnemonicPath(deviceType, holderStructure);

        return getStructureElement(
                deviceType.getUuid(),
                Type.DEVICETYPE,
                deviceType.getParentUuid(),
                deviceType.getName(), deviceType.getMnemonic(), mnemonicPath, 3,
                deviceType.getDescription(), deviceType.getStatus(), deviceType.isLatest(), deviceType.isDeleted(),
                deviceType.getProcessed(), deviceType.getProcessedBy(), deviceType.getProcessedComment());
    }

    /**
     * Populate and return structure element.
     *
     * @param uuid uuid
     * @param type type
     * @param parent parent uuid
     * @param name name
     * @param mnemonic mnemonic
     * @param mnemonicPath mnemonic path
     * @param level level
     * @param description description
     * @param status status
     * @param latest latest
     * @param deleted deleted
     * @param when when
     * @param who who
     * @param comment comment
     * @return structure element
     */
    protected static StructureElement getStructureElement(
            UUID uuid,
            Type type,
            UUID parent,
            String name, String mnemonic, String mnemonicPath, Integer level,
            String description, Status status, Boolean latest, Boolean deleted,
            Date when, String who, String comment) {

        return new StructureElement(
                uuid, type, parent, name, mnemonic,
                description, comment,
                mnemonicPath, level,
                status, latest, deleted,
                when, who);
    }

    private static boolean hasSameContent(StructureElement structureElement, NameStructure nameStructure) {
        // StructureElement
        //     x uuid
        //     x description
        //     x status
        //     x latest
        //     x deleted
        // NameStructure
        //     x uuid
        //     x description
        //     x status
        //     x latest
        //     x deleted

        if (structureElement == null && nameStructure == null)
            return true;
        if (structureElement == null)
            return false;
        if (nameStructure == null)
            return false;

        if (structureElement.getUuid() == null) {
            if (nameStructure.getUuid() != null)
                return false;
        } else if (!structureElement.getUuid().equals(nameStructure.getUuid()))
            return false;
        if (structureElement.getDescription() == null) {
            if (nameStructure.getDescription() != null)
                return false;
        } else if (!structureElement.getDescription().equals(nameStructure.getDescription()))
            return false;
        if (structureElement.getStatus() == null) {
            if (nameStructure.getStatus() != null)
                return false;
        } else if (!structureElement.getStatus().equals(nameStructure.getStatus()))
            return false;
        if (structureElement.isLatest() == null) {
            if (nameStructure.isLatest() != null)
                return false;
        } else if (!structureElement.isLatest().equals(nameStructure.isLatest()))
            return false;
        if (structureElement.isDeleted() == null) {
            if (nameStructure.isDeleted() != null)
                return false;
        } else if (!structureElement.isDeleted().equals(nameStructure.isDeleted()))
            return false;

        return true;
    }

    private static boolean hasSameContent(StructureElement structureElement, Structure structure) {
        // StructureElement
        //     x name
        //     x mnemonic
        // Structure
        //     x name
        //     x mnemonic

        if (!hasSameContent(structureElement, (NameStructure) structure))
            return false;

        if (structureElement == null)
            return false;

        if (structureElement.getName() == null) {
            if (structure.getName() != null)
                return false;
        } else if (!structureElement.getName().equals(structure.getName()))
            return false;
        if (structureElement.getMnemonic() == null) {
            if (structure.getMnemonic() != null)
                return false;
        } else if (!structureElement.getMnemonic().equals(structure.getMnemonic()))
            return false;

        return true;
    }

    public static boolean hasSameContent(StructureElement structureElement, SystemGroup systemGroup) {
        // StructureElement
        //     x type
        // SystemGroup

        if (!hasSameContent(structureElement, (Structure) systemGroup))
            return false;

        return Type.SYSTEMGROUP.equals(structureElement.getType());
    }

    public static boolean hasSameContent(StructureElement structureElement, System system) {
        // StructureElement
        //     x type
        //     x parent
        // System
        //     x parent_uuid

        if (!hasSameContent(structureElement, (Structure) system))
            return false;

        if (!Type.SYSTEM.equals(structureElement.getType())) {
            return false;
        }
        if (structureElement.getParent() == null) {
            if (system.getParentUuid() != null)
                return false;
        } else if (!structureElement.getParent().equals(system.getParentUuid()))
            return false;

        return true;
    }

    public static boolean hasSameContent(StructureElement structureElement, Subsystem subsystem) {
        // StructureElement
        //     x type
        //     x parent
        // Subsystem
        //     x parent_uuid

        if (!hasSameContent(structureElement, (Structure) subsystem))
            return false;

        if (!Type.SUBSYSTEM.equals(structureElement.getType())) {
            return false;
        }
        if (structureElement.getParent() == null) {
            if (subsystem.getParentUuid() != null)
                return false;
        } else if (!structureElement.getParent().equals(subsystem.getParentUuid()))
            return false;

        return true;
    }

    public static boolean hasSameContent(StructureElement structureElement, Discipline discipline) {
        // StructureElement
        //     x type
        // Discipline

        if (!hasSameContent(structureElement, (Structure) discipline))
            return false;

        return Type.DISCIPLINE.equals(structureElement.getType());
    }

    public static boolean hasSameContent(StructureElement structureElement, DeviceGroup deviceGroup) {
        // StructureElement
        //     x type
        //     x parent
        // Subsystem
        //     x parent_uuid

        if (!hasSameContent(structureElement, (Structure) deviceGroup))
            return false;

        if (!Type.DEVICEGROUP.equals(structureElement.getType())) {
            return false;
        }
        if (structureElement.getParent() == null) {
            if (deviceGroup.getParentUuid() != null)
                return false;
        } else if (!structureElement.getParent().equals(deviceGroup.getParentUuid()))
            return false;

        return true;
    }

    public static boolean hasSameContent(StructureElement structureElement, DeviceType deviceType) {
        // StructureElement
        //     x type
        //     x parent
        // DeviceType
        //     x parent_uuid

         if (!hasSameContent(structureElement, (Structure) deviceType))
             return false;

         if (!Type.DEVICETYPE.equals(structureElement.getType())) {
             return false;
         }
         if (structureElement.getParent() == null) {
             if (deviceType.getParentUuid() != null)
                 return false;
         } else if (!structureElement.getParent().equals(deviceType.getParentUuid()))
             return false;

         return true;
    }

    // ----------------------------------------------------------------------------------------------------

    /**
     * Convert structure element command for create to structure element command.
     *
     * @param command structure element command for create
     * @return structure element command
     */
    public static StructureElementCommand convertCommandCreate2Command(StructureElementCommandCreate command) {
        return new StructureElementCommand(
                null,
                command.getType(),
                command.getParent(),
                command.getName(),
                command.getMnemonic(),
                command.getDescription(),
                command.getComment());
    }

    /**
     * Convert list of structure element commands for create to list of structure element commands.
     *
     * @param commands list of structure element commands for create
     * @return list of structure element commands
     */
    public static List<StructureElementCommand> convertCommandCreate2Command(List<StructureElementCommandCreate> commands) {
        List<StructureElementCommand> structureElementCommands = Lists.newArrayList();
        for (StructureElementCommandCreate command : commands) {
            structureElementCommands.add(StructureElementUtil.convertCommandCreate2Command(command));
        }
        return structureElementCommands;
    }

    /**
     * Convert structure element command for update to structure element command.
     *
     * @param command structure element command for update
     * @return structure element command
     */
    public static StructureElementCommand convertCommandUpdate2Command(StructureElementCommandUpdate command) {
        return new StructureElementCommand(
                command.getUuid(),
                command.getType(),
                command.getParent(),
                command.getName(),
                command.getMnemonic(),
                command.getDescription(),
                command.getComment());
    }

    /**
     * Convert list of structure element commands for update to list of structure element commands.
     *
     * @param commands list of structure element commands for update
     * @return list of structure element commands
     */
    public static List<StructureElementCommand> convertCommandUpdate2Command(List<StructureElementCommandUpdate> commands) {
        List<StructureElementCommand> structureElementCommands = Lists.newArrayList();
        for (StructureElementCommandUpdate command : commands) {
            structureElementCommands.add(StructureElementUtil.convertCommandUpdate2Command(command));
        }
        return structureElementCommands;
    }

    /**
     * Convert structure element command for confirm to structure element command.
     *
     * @param command structure element command for confirm
     * @return structure element command
     */
    public static StructureElementCommand convertCommandConfirm2Command(StructureElementCommandConfirm command) {
        return new StructureElementCommand(
                command.getUuid(),
                command.getType(),
                null,
                null,
                null,
                null,
                command.getComment());
    }

    /**
     * Convert list of structure element commands for confirm to list of structure element commands.
     *
     * @param commands list of structure element commands for confirm
     * @return list of structure element commands
     */
    public static List<StructureElementCommand> convertCommandConfirm2Command(List<StructureElementCommandConfirm> commands) {
        List<StructureElementCommand> structureElementCommands = Lists.newArrayList();
        for (StructureElementCommandConfirm command : commands) {
            structureElementCommands.add(StructureElementUtil.convertCommandConfirm2Command(command));
        }
        return structureElementCommands;
    }

    // ----------------------------------------------------------------------------------------------------

    /**
     * Convert structure element command to structure element command for create.
     *
     * @param command structure element command
     * @return structure element command for create
     */
    public static StructureElementCommandCreate convertCommand2CommandCreate(StructureElementCommand command) {
        return new StructureElementCommandCreate(
                command.getType(),
                command.getParent(),
                command.getName(),
                command.getMnemonic(),
                command.getDescription(),
                command.getComment());
    }

    /**
     * Convert array of structure element commands to array of structure element commands for create.
     *
     * @param commands array of structure element commands
     * @return array of structure element commands for create
     */
    public static StructureElementCommandCreate[] convertCommand2CommandCreate(StructureElementCommand[] commands) {
        StructureElementCommandCreate[] structureElementCommands = new StructureElementCommandCreate[commands.length];
        for (int i=0; i<commands.length; i++) {
            structureElementCommands[i] = StructureElementUtil.convertCommand2CommandCreate(commands[i]);

        }
        return structureElementCommands;
    }

    /**
     * Convert structure element command to structure element command for update.
     *
     * @param command structure element command
     * @return structure element command for update
     */
    public static StructureElementCommandUpdate convertCommand2CommandUpdate(StructureElementCommand command) {
        return new StructureElementCommandUpdate(
                command.getUuid(),
                command.getType(),
                command.getParent(),
                command.getName(),
                command.getMnemonic(),
                command.getDescription(),
                command.getComment());
    }

    /**
     * Convert array of structure element commands to array of structure element commands for update.
     *
     * @param commands array of structure element commands
     * @return array of structure element commands for update
     */
    public static StructureElementCommandUpdate[] convertCommand2CommandUpdate(StructureElementCommand[] commands) {
        StructureElementCommandUpdate[] structureElementCommands = new StructureElementCommandUpdate[commands.length];
        for (int i=0; i<commands.length; i++) {
            structureElementCommands[i] = StructureElementUtil.convertCommand2CommandUpdate(commands[i]);

        }
        return structureElementCommands;
    }

    /**
     * Convert structure element command to structure element command for confirm.
     *
     * @param command structure element command
     * @return structure element command for confirm
     */
    public static StructureElementCommandConfirm convertCommand2CommandConfirm(StructureElementCommand command) {
        return new StructureElementCommandConfirm(
                command.getUuid(),
                command.getType(),
                command.getComment());
    }

    /**
     * Convert array of structure element commands to array of structure element commands for confirm.
     *
     * @param commands array of structure element commands
     * @return array of structure element commands for confirm
     */
    public static StructureElementCommandConfirm[] convertCommand2CommandConfirm(StructureElementCommand[] commands) {
        StructureElementCommandConfirm[] structureElementCommands = new StructureElementCommandConfirm[commands.length];
        for (int i=0; i<commands.length; i++) {
            structureElementCommands[i] = StructureElementUtil.convertCommand2CommandConfirm(commands[i]);
        }
        return structureElementCommands;
    }

    // ----------------------------------------------------------------------------------------------------

    /**
     * Convert structure element to structure element command for create.
     *
     * @param element structure element
     * @return structure element command for create
     */
    public static StructureElementCommandCreate convertElement2CommandCreate(StructureElement element) {
        return new StructureElementCommandCreate(
                element.getType(),
                element.getParent(),
                element.getName(),
                element.getMnemonic(),
                element.getDescription(),
                element.getComment());
    }

    /**
     * Convert array of structure elements to array of structure element commands for create.
     *
     * @param elements array of structure elements
     * @return array of structure element commands for create
     */
    public static StructureElementCommandCreate[] convertElement2CommandCreate(StructureElement[] elements) {
        StructureElementCommandCreate[] structureElementCommands = new StructureElementCommandCreate[elements.length];
        for (int i=0; i<elements.length; i++) {
            structureElementCommands[i] = convertElement2CommandCreate(elements[i]);
        }
        return structureElementCommands;
    }

    /**
     * Convert structure element to structure element command for update.
     *
     * @param element structure element
     * @return structure element command for update
     */
    public static StructureElementCommandUpdate convertElement2CommandUpdate(StructureElement element) {
        return new StructureElementCommandUpdate(
                element.getUuid(),
                element.getType(),
                element.getParent(),
                element.getName(),
                element.getMnemonic(),
                element.getDescription(),
                element.getComment());
    }

    /**
     * Convert array of structure elements to array of structure element commands for update.
     *
     * @param elements array of structure elements
     * @return array of structure element commands for update
     */
    public static StructureElementCommandUpdate[] convertElement2CommandUpdate(StructureElement[] elements) {
        StructureElementCommandUpdate[] structureElementCommands = new StructureElementCommandUpdate[elements.length];
        for (int i=0; i<elements.length; i++) {
            structureElementCommands[i] =  convertElement2CommandUpdate(elements[i]);
        }
        return structureElementCommands;
    }

    /**
     * Convert structure element to structure element command for confirm.
     *
     * @param element structure element
     * @return structure element command for confirm
     */
    public static StructureElementCommandConfirm convertElement2CommandConfirm(StructureElement element) {
        return new StructureElementCommandConfirm(
                element.getUuid(),
                element.getType(),
                element.getComment());
    }

    /**
     * Convert array of structure elements to array of structure element commands for confirm.
     *
     * @param elements array of structure elements
     * @return array of structure element commands for confirm
     */
    public static StructureElementCommandConfirm[] convertElement2CommandConfirm(StructureElement[] elements) {
        StructureElementCommandConfirm[] structureElementCommands = new StructureElementCommandConfirm[elements.length];
        for (int i=0; i<elements.length; i++) {
            structureElementCommands[i] = convertElement2CommandConfirm(elements[i]);
        }
        return structureElementCommands;
    }


}
